# coding: utf8
"""Py2/3 compatibility"""
import sys

is_python2 = int(sys.version[0]) == 2
is_windows = sys.platform.startswith('win')
is_linux = sys.platform.startswith('linux')
is_osx = sys.platform == 'darwin'

try:
    unicode_ = unicode  # Python 2
    bytes_ = str
    string_types = (str, unicode)
    chr_ = unichr

    def unicode_to_bytes(s, encoding='utf8', errors='strict'):
        return s.encode(encoding=encoding, errors=errors)

    def bytes_to_unicode(b, encoding='utf8', errors='strict'):
        return unicode_(b, encoding=encoding, errors=errors)

except NameError:
    unicode_ = str      # Python 3
    bytes_ = bytes
    string_types = (bytes, str)
    chr_ = chr

    def unicode_to_bytes(s, encoding='utf8', errors='strict'):
        return s.encode(encoding=encoding, errors=errors)

    def bytes_to_unicode(b, encoding='utf8', errors='strict'):
        return b.decode(encoding=encoding, errors=errors)



# coding: utf8
"""Conll Evaluation - Scoring"""
from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

import os
import subprocess
import io
#import concurrent.futures
import pickle

from torch.autograd import Variable
from torch.utils.data import DataLoader

#from algorithm import Coref
from neuralcoref.conllparser import FEATURES_NAMES
from neuralcoref.dataset import NCBatchSampler, padder_collate
from neuralcoref.compat import unicode_

PACKAGE_DIRECTORY = os.path.dirname(os.path.abspath(__file__))

OUT_PATH = os.path.join(PACKAGE_DIRECTORY, "test_corefs.txt") #fernandes.txt")#
ALL_MENTIONS_PATH = os.path.join(PACKAGE_DIRECTORY, "test_mentions.txt")
#KEY_PATH = os.path.join(PACKAGE_DIRECTORY, "conll-2012-test-test-key.txt")
SCORING_SCRIPT = os.path.join(PACKAGE_DIRECTORY, "scorer_wrapper.pl")

METRICS = ['muc', 'bcub', 'ceafm', 'ceafe', 'blanc']
CONLL_METRICS = ['muc', 'bcub', 'ceafe']

class ConllEvaluator(object):
    def __init__(self, model, dataset, test_data_path, test_key_file, embed_path,
                 args):
        """ Evaluate the pytorch model that is currently being build
            We take the embedding vocabulary currently being trained
        """
        self.test_key_file = test_key_file
        self.cuda = args.cuda
        self.model = model
        batch_sampler = NCBatchSampler(dataset.mentions_pair_length,
                                       batchsize=args.batchsize, shuffle=False)
        self.dataloader = DataLoader(dataset,
                                     collate_fn=padder_collate,
                                     batch_sampler=batch_sampler,
                                     num_workers=args.numworkers,
                                     pin_memory=args.cuda)
        self.mentions_idx, self.n_pairs = batch_sampler.get_batch_info()
        self.load_meta(test_data_path)

    def load_meta(self, test_data_path):
        # Load meta files
        datas = {}
        if not os.listdir(test_data_path):
            raise ValueError("Empty test_data_path")
        bin_files_found = False
        print("Reading ", end='')
        for file_name in os.listdir(test_data_path):
            if '.bin' not in file_name:
                continue
            bin_files_found = True
            print(file_name, end=', ')
            with open(test_data_path + file_name, 'rb') as f:
                datas[file_name.split(u'.')[0]] = pickle.load(f)
        if not bin_files_found:
            raise ValueError("Can't find bin files in {}".format(test_data_path))
        print("Done")
        self.m_loc = datas[FEATURES_NAMES[9]]
        self.tokens = datas[FEATURES_NAMES[10]]
        self.lookup = datas[FEATURES_NAMES[11]]
        self.docs = datas[FEATURES_NAMES[12]]
        self.flat_m_idx = list((doc_i, m_i) for doc_i, l in enumerate(self.m_loc) for m_i in range(len(l)))

    ###########################
    #### CLUSTER FUNCTIONS ####
    ###########################

    def _prepare_clusters(self):
        '''
        Clean up and prepare one cluster for each mention
        '''
        self.mention_to_cluster = list(list(range(len(doc_mentions))) for doc_mentions in self.m_loc)
        self.clusters = list(dict((i, [i]) for i in doc_mentions) for doc_mentions in self.mention_to_cluster)

    def _merge_coreference_clusters(self, ant_flat_idx, mention_flat_idx):
        '''
        Merge two clusters together
        '''
        doc_idx, ant_idx = self.flat_m_idx[ant_flat_idx]
        doc_idx2, mention_idx = self.flat_m_idx[mention_flat_idx]
        assert doc_idx2 == doc_idx
        if self.mention_to_cluster[doc_idx][ant_idx] == self.mention_to_cluster[doc_idx][mention_idx]:
            return
        remove_id = self.mention_to_cluster[doc_idx][ant_idx]
        keep_id = self.mention_to_cluster[doc_idx][mention_idx]
        for idx in self.clusters[doc_idx][remove_id]:
            self.mention_to_cluster[doc_idx][idx] = keep_id
            self.clusters[doc_idx][keep_id].append(idx)
        del self.clusters[doc_idx][remove_id]

    def remove_singletons_clusters(self, debug=False):
        for doc_idx in range(len(self.docs)):
            remove_id = []
            kept = False
            for key, mentions in self.clusters[doc_idx].items():
                if len(mentions) == 1:
                    remove_id.append(key)
                    self.mention_to_cluster[doc_idx][key] = None
                else:
                    kept = True
                    if debug:
                        l = list(self.m_loc[doc_idx][m][3] for m in mentions)
                        print("Cluster found", key)
                        print("Corefs:", "|".join(str(self.docs[doc_idx]['mentions'][m_idx]) \
                              + " (" + str(m_idx) + ")" for m_idx in l))
            if not kept:
                print("❄️ No coreference found")
            for rem in remove_id:
                del self.clusters[doc_idx][rem]

    def display_clusters(self, doc_idx=None):
        '''
        Print clusters informations
        '''
        doc_it = range(len(self.docs)) if doc_idx is None else [doc_idx]
        for d_i in doc_it:
            print("Clusters in doc:", doc_it, self.docs[d_i]['name'], self.docs[d_i]['part'])
            print(self.clusters[d_i])
            for key, mentions in self.clusters[d_i].items():
                l = list(self.m_loc[d_i][m][3] for m in mentions)
                print("cluster", key, "(", ", ".join(self.docs[d_i]['mentions'][m_idx] for m_idx in l), ")")

    ########################
    #### MAIN FUNCTIONS ####
    ########################
    def get_max_score(self, batch, debug=False):
        inputs, mask = batch
        inputs = tuple(Variable(i, volatile=True) for i in inputs)
        if self.cuda:
            inputs = tuple(i.cuda() for i in inputs)
            mask = mask.cuda()
        self.model.eval()
        scores = self.model.forward(inputs, concat_axis=1).data
        scores.masked_fill_(mask, -float('Inf'))
        _, max_idx = scores.max(dim=1) # We may want to weight the single score with coref.greedyness
        if debug:
            print("Max_idx", max_idx)
        return scores.cpu().numpy(), max_idx.cpu().numpy()

    def test_model(self):
        print("🌋 Test evaluator / print all mentions")
        self.build_test_file(out_path=ALL_MENTIONS_PATH, print_all_mentions=True)
        self.get_score(file_path=ALL_MENTIONS_PATH)

    def build_test_file(self, out_path=OUT_PATH, remove_singleton=True, print_all_mentions=False, debug=None):
        """ Build a test file to supply to the coreference scoring perl script
        """
        print("🌋 Building test file")
        self._prepare_clusters()
        self.dataloader.dataset.no_targets = True
        if not print_all_mentions:
            print("🌋 Build coreference clusters")
            cur_m = 0
            for sample_batched, mentions_idx, n_pairs_l in zip(self.dataloader, self.mentions_idx, self.n_pairs):
                scores, max_i = self.get_max_score(sample_batched)
                for m_idx, ind, n_pairs in zip(mentions_idx, max_i, n_pairs_l):
                    if ind < n_pairs : # the single score is not the highest, we have a match !
                        prev_idx = m_idx - n_pairs + ind
                        if debug is not None and (debug == -1 or debug == prev_idx or debug == m_idx):
                            m1_doc, m1_idx = self.flat_m_idx[m_idx]
                            m1 = self.docs[m1_doc]['mentions'][m1_idx]
                            m2_doc, m2_idx = self.flat_m_idx[prev_idx]
                            m2 = self.docs[m2_doc]['mentions'][m2_idx]
                            print("We have a match between:", m1, "(" + str(m1_idx) + ")", "and:", m2, "(" + str(m2_idx) + ")")
                        self._merge_coreference_clusters(prev_idx, m_idx)
            if remove_singleton:
                self.remove_singletons_clusters()
        self.dataloader.dataset.no_targets = False

        print("🌋 Construct test file")
        out_str = ""
        for doc, d_tokens, d_lookup, d_m_loc, d_m_to_c in zip(self.docs, self.tokens, self.lookup, self.m_loc, self.mention_to_cluster):
            out_str += u"#begin document (" + doc['name'] + u"); part " + doc['part'] + u"\n"
            for utt_idx, (c_tokens, c_lookup) in enumerate(zip(d_tokens, d_lookup)):
                for i, (token, lookup) in enumerate(zip(c_tokens, c_lookup)):
                    out_coref = u""
                    for m_str, mention, mention_cluster in zip(doc['mentions'], d_m_loc, d_m_to_c):
                        m_start, m_end, m_utt, m_idx, m_doc = mention
                        if mention_cluster is None:
                            pass
                        elif m_utt == utt_idx:
                            if m_start in lookup:
                                out_coref += u"|" if out_coref else u""
                                out_coref += u"(" + unicode_(mention_cluster)
                                if (m_end - 1) in lookup:
                                    out_coref += u")"
                                else:
                                    out_coref += u""
                            elif (m_end - 1) in lookup:
                                out_coref += u"|" if out_coref else u""
                                out_coref += unicode_(mention_cluster) + u")"
                    out_line = doc['name'] + u" " + doc['part'] + u" " + unicode_(i) \
                               + u" " + token + u" "
                    out_line += u"-" if len(out_coref) == 0 else out_coref
                    out_str += out_line + u"\n"
                out_str += u"\n"
            out_str += u"#end document\n"

        # Write test file
        print("Writing in", out_path)
        with io.open(out_path, 'w', encoding='utf-8') as out_file:
            out_file.write(out_str)

    def get_score(self, file_path=OUT_PATH, debug=False):
        """ Call the coreference scoring perl script on the created test file
        """
        print("🌋 Computing score")
        score = {}
        ident = None
        for metric_name in CONLL_METRICS:
            if debug: print("Computing metric:", metric_name)
            try:
                scorer_out = subprocess.check_output(["perl",
                                                      SCORING_SCRIPT,
                                                      metric_name,
                                                      self.test_key_file,
                                                      file_path], stderr=subprocess.STDOUT, encoding='utf-8')
            except subprocess.CalledProcessError as err:
                print("Error during the scoring")
                print(err)
                print(err.output)
                raise
            if debug: print("scorer_out", scorer_out)
            value, ident = scorer_out.split(u"\n")[-2], scorer_out.split(u"\n")[-1]
            if debug: print("value", value, "identification", ident)
            NR, DR, NP, DP = [float(x) for x in value.split(u" ")]
            ident_NR, ident_DR, ident_NP, ident_DP = [float(x) for x in ident.split(u" ")]
            precision = NP/DP if DP else 0
            recall = NR/DR if DR else 0
            F1 = 2 * precision * recall / (precision + recall) if precision + recall > 0 else 0
            ident_precision = ident_NP/ident_DP if ident_DP else 0
            ident_recall = ident_NR/ident_DR if ident_DR else 0
            ident_F1 = 2 * ident_precision * ident_recall / (ident_precision + ident_recall) if ident_precision + ident_recall > 0 else 0
            score[metric_name] = (precision, recall, F1)
            ident = (ident_precision, ident_recall, ident_F1, ident_NR, ident_DR, ident_NP, ident_DP)
        F1_conll = sum([score[metric][2] for metric in CONLL_METRICS])/len(CONLL_METRICS)
        print("Mention identification recall", ident[1], "<= Detected mentions", ident[3], "True mentions", ident[4])
        print("Scores", score)
        print("F1_conll", F1_conll)
        return score, F1_conll, ident



# coding: utf8
"""Conll training algorithm"""
from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

import os
import io
import numpy as np

import torch
import torch.utils.data

from torch.utils.data.sampler import Sampler
from torch.utils.data import Dataset

from neuralcoref.utils import (encode_distance, BATCH_SIZE_PATH, SIZE_FP,
                               SIZE_FP_COMPRESSED, SIZE_FS, SIZE_FS_COMPRESSED,
                               SIZE_GENRE, SIZE_PAIR_IN, SIZE_SINGLE_IN)
from neuralcoref.conllparser import FEATURES_NAMES

def load_embeddings_from_file(name):
    print("loading", name+"_embeddings.npy")
    embed = torch.from_numpy(np.load(name+"_embeddings.npy")).float()
    print(embed.size())
    print("loading", name+"_vocabulary.txt")
    with io.open(name+"_vocabulary.txt", 'r', encoding='utf-8') as f:
        voc = [line.strip() for line in f]
    return embed, voc

class NCDataset(Dataset):
    def __init__(self, data_path, params, no_targets=False):
        print("🏝 Loading Dataset at", data_path)
        self.costs = params.costs
        self.no_targets = no_targets
        # Load files
        datas = {}
        if not os.listdir(data_path):
            raise ValueError("Empty data_path")
        numpy_files_found = False
        print("Reading ", end='')
        for file_name in os.listdir(data_path):
            if not '.npy' in file_name:
                continue
            numpy_files_found = True
            print(file_name, end=', ')
            datas[file_name.split(u'.')[0]] = np.load(data_path + file_name)
        if not numpy_files_found:
            raise ValueError("Can't find numpy files in {}".format(data_path))

        # Gather arrays in two lists of tuples for mention and pairs
        self.mentions = list(zip(*(arr for key, arr in sorted(datas.items()) if key.startswith(u"mentions"))))
        self.pairs = list(zip(*(arr for key, arr in sorted(datas.items()) if key.startswith(u"pairs"))))
        self.mentions_pair_length = datas[FEATURES_NAMES[2]]
        assert [arr.shape[0] for arr in self.mentions[0]] == [6, 1, 1, 1, 250, 8] # Cf order of FEATURES_NAMES in conllparser.py
        assert [arr.shape[0] for arr in self.pairs[0]] == [1, 9, 1] # Cf order of FEATURES_NAMES in conllparser.py

    def __len__(self):
        return len(self.mentions)

    def __getitem__(self, mention_idx, debug=False):
        """
        Return:
            Definitions:
                P is the number of antecedent per mention (number of pairs for the mention)
                S = 250 is the size of the span vector (averaged word embeddings)
                W = 8 is the number of words in a mention (tuned embeddings)
                Fp = 70 is the number of features for a pair of mention
                Fs = 24 is the number of features of a single mention

            if there are some pairs:
                inputs = (spans, words, features, ant_spans, ant_words, ana_spans, ana_words, pairs_features)
                targets = (labels, costs, true_ants, false_ants)
            else:
                inputs = (spans, words, features)
                targets = (labels, costs, true_ants)

            inputs: Tuple of
                spans => (S,)
                words => (W,)
                features => (Fs,)
                + if there are potential antecedents (P > 0):
                    ant_spans => (P, S) or nothing if no pairs
                    ant_words => (P, W) or nothing if no pairs
                    ana_spans => (P, S) or nothing if no pairs
                    ana_words => (P, W) or nothing if no pairs
                    pair_features => (P, Fp) or nothing if no pairs

            targets: Tuple of
                labels => (P+1,)
                costs => (P+1,)
                true_ant => (P+1,)
                + if there are potential antecedents (P > 0):
                    false_ant => (P+1,)

        """
        features_raw, label, pairs_length, pairs_start_index, spans, words = self.mentions[mention_idx]
        pairs_start_index = np.asscalar(pairs_start_index)
        pairs_length = np.asscalar(pairs_length)

        # Build features array (float) from raw features (int)
        assert features_raw.shape[0] == SIZE_FS_COMPRESSED
        features = np.zeros((SIZE_FS,))
        features[features_raw[0]] = 1
        features[4:15] = encode_distance(features_raw[1])
        features[15] = features_raw[2].astype(float) / features_raw[3].astype(float)
        features[16] = features_raw[4]
        features[features_raw[5] + 17] = 1

        if pairs_length == 0:
            spans = torch.from_numpy(spans).float()
            words = torch.from_numpy(words)
            features = torch.from_numpy(features).float()
            inputs = (spans, words, features)
            if self.no_targets:
                return inputs
            true_ant = torch.zeros(1).long() # zeros = indices of true ant
            costs = torch.from_numpy((1 - label) * self.costs['FN']).float()
            label = torch.from_numpy(label).float()
            targets = (label, costs, true_ant)
            if debug:
                print("inputs shapes: ", [a.size() for a in inputs])
                print("targets shapes: ", [a.size() for a in targets])
            return inputs, targets

        start = pairs_start_index
        end = pairs_start_index + pairs_length
        pairs = self.pairs[start:end]
        assert len(pairs) == pairs_length
        assert len(pairs[0]) == 3 # pair[i] = (pairs_ant_index, pairs_features, pairs_labels)
        pairs_ant_index, pairs_features_raw, pairs_labels = list(zip(*pairs))

        pairs_features_raw = np.stack(pairs_features_raw)
        pairs_labels = np.squeeze(np.stack(pairs_labels), axis=1)

        # Build pair features array (float) from raw features (int)
        assert pairs_features_raw[0, :].shape[0] == SIZE_FP_COMPRESSED
        pairs_features = np.zeros((len(pairs_ant_index), SIZE_FP))
        pairs_features[:, 0:6] = pairs_features_raw[:, 0:6]
        pairs_features[:, 6:17] = encode_distance(pairs_features_raw[:, 6])
        pairs_features[:, 17:28] = encode_distance(pairs_features_raw[:, 7])
        pairs_features[:, 28] = pairs_features_raw[:, 8]
        # prepare antecent features
        ant_features_raw = np.concatenate([self.mentions[np.asscalar(idx)][0][np.newaxis, :] for idx in pairs_ant_index])
        ant_features = np.zeros((pairs_length, SIZE_FS-SIZE_GENRE))
        ant_features[:, ant_features_raw[:, 0]] = 1
        ant_features[:, 4:15] = encode_distance(ant_features_raw[:, 1])
        ant_features[:, 15] = ant_features_raw[:, 2].astype(float) / ant_features_raw[:, 3].astype(float)
        ant_features[:, 16] = ant_features_raw[:, 4]
        pairs_features[:, 29:46] = ant_features
        # Here we keep the genre 
        ana_features = np.tile(features, (pairs_length, 1))
        pairs_features[:, 46:] = ana_features

        ant_spans = np.concatenate([self.mentions[np.asscalar(idx)][4][np.newaxis, :] for idx in pairs_ant_index])
        ant_words = np.concatenate([self.mentions[np.asscalar(idx)][5][np.newaxis, :] for idx in pairs_ant_index])
        ana_spans = np.tile(spans, (pairs_length, 1))
        ana_words = np.tile(words, (pairs_length, 1))
        ant_spans = torch.from_numpy(ant_spans).float()
        ant_words = torch.from_numpy(ant_words)
        ana_spans = torch.from_numpy(ana_spans).float()
        ana_words = torch.from_numpy(ana_words)
        pairs_features = torch.from_numpy(pairs_features).float()

        labels_stack = np.concatenate((pairs_labels, label), axis=0)
        assert labels_stack.shape == (pairs_length + 1,)
        labels = torch.from_numpy(labels_stack).float()

        spans = torch.from_numpy(spans).float()
        words = torch.from_numpy(words)
        features = torch.from_numpy(features).float()

        inputs = (spans, words, features,
                  ant_spans, ant_words,
                  ana_spans, ana_words,
                  pairs_features)

        if self.no_targets:
            return inputs

        if label == 0:
            costs = np.concatenate((self.costs['WL'] * (1 - pairs_labels), [self.costs['FN']])) # Inverse labels: 1=>0, 0=>1
        else:
            costs = np.concatenate((self.costs['FL'] * np.ones_like(pairs_labels), [0]))
        assert costs.shape == (pairs_length + 1,)
        costs = torch.from_numpy(costs).float()

        true_ants_unpad = np.flatnonzero(labels_stack)
        if len(true_ants_unpad) == 0:
            raise ValueError("Error: no True antecedent for mention")
        true_ants = np.pad(true_ants_unpad, (0, len(pairs_labels) + 1 - len(true_ants_unpad)), 'edge')
        assert true_ants.shape == (pairs_length + 1,)
        true_ants = torch.from_numpy(true_ants).long()

        false_ants_unpad = np.flatnonzero(1 - labels_stack)
        assert len(false_ants_unpad) != 0
        false_ants = np.pad(false_ants_unpad, (0, len(pairs_labels) + 1 - len(false_ants_unpad)), 'edge')
        assert false_ants.shape == (pairs_length + 1,)
        false_ants = torch.from_numpy(false_ants).long()

        targets = (labels, costs, true_ants, false_ants)
        if debug:
            print("Mention", mention_idx)
            print("inputs shapes: ", [a.size() for a in inputs])
            print("targets shapes: ", [a.size() for a in targets])
        return inputs, targets

class NCBatchSampler(Sampler):
    """A Batch sampler to group mentions in batches with close number of pairs to be padded together
    """

    def __init__(self, mentions_pairs_length, batchsize=600,
                 shuffle=False, debug=False):
        """ Create and feed batches of mentions having close number of antecedents
            The batch are padded and collated by the padder_collate function
        
        # Arguments:
            mentions_pairs_length array of shape (N, 1): list/array of the number of pairs for each mention
            batchsize: Number of pairs of each batch will be capped at this
        """
        self.shuffle = shuffle
        num_mentions = len(mentions_pairs_length)
        mentions_lengths = np.concatenate([mentions_pairs_length, np.arange(0, num_mentions, 1, dtype=int)[:, np.newaxis]], axis=1)
        sorted_lengths = mentions_lengths[mentions_lengths[:, 0].argsort()]
        print("Preparing batches 📚")

        self.batches = []
        self.batches_pairs = []
        self.batches_size = []
        batch = []
        n_pairs = []
        num = 0
        for length, mention_idx in sorted_lengths:
            if num > batchsize or (num == len(batch) and length != 0): # We keep the no_pairs batches pure
                if debug: print("Added batch number", len(self.batches), 
                                "with", len(batch), "mentions and", num, "pairs")
                self.batches.append(batch)
                self.batches_size.append(num) # We don't count the max 7 additional mentions that are repeated
                self.batches_pairs.append(n_pairs)

                # Start a new batch
                batch = [mention_idx]
                n_pairs = [length]
                num = length + 1 # +1 since we also have the single mention to add to the number of pairs
            else:
                num += length + 1
                batch.append(mention_idx)
                n_pairs.append(length)

        # Complete and store the last batch
        if debug: print("Added batch number", len(self.batches),"with", len(batch), "mentions and", num, "pairs")
        self.batches.append(batch)
        self.batches_size.append(num)
        self.batches_pairs.append(n_pairs)
        self.n_pairs = sum(sum(p) for p in self.batches_pairs)
        self.n_mentions = sum(len(b) for b in self.batches)
        self.n_batches = len(self.batches)
        self.pairs_per_batch = float(self.n_pairs) / self.n_batches
        self.mentions_per_batch = float(self.n_mentions) / self.n_batches
        print("Dataset has:", self.n_batches, "batches,", self.n_mentions, "mentions,", self.n_pairs, "pairs")

    def get_batch_info(self):
        return self.batches, self.batches_pairs

    def save_batch_sizes(self, save_file=BATCH_SIZE_PATH, debug=False):
        print("🌋 Saving sizes of batches")
        with io.open(save_file, "w", encoding='utf-8') as f:
            if debug: print("Batch sizes saved in", save_file)
            for batch, size in zip(self.batches, self.batches_size):
                out_str = str(len(batch)) + "\t" + str(size) + "\n"
                f.write(out_str)

    def __iter__(self):
        if self.shuffle:
            np.random.shuffle(self.batches)
        for batch in self.batches:
            yield batch

    def __len__(self):
        return self.n_batches

def padder_collate(batch, debug=False):
    """ Puts each data field into a tensor with outer dimension batch size
        Pad variable length input tensors and add a weight tensor to the target 
    """
    transposed_inputs = tuple(zip(*batch))
    if len(transposed_inputs) == 2:
        inputs, targets = transposed_inputs
        transposed_inputs = tuple(zip(*inputs))
        transposed_targets = tuple(zip(*targets))
    else:
        transposed_targets = None

    max_pairs = max(len(t) for t in transposed_inputs[3]) if len(transposed_inputs) == 8 else 0 # Get max nb of pairs (batch are sorted by nb of pairs)
    if max_pairs > 0:
        out_inputs = []
        out_targets = []
        for t_inp in transposed_inputs:
            if len(t_inp[0].shape) == 2:
                out_inputs.append(torch.stack([torch.cat([t, t.new(max_pairs - len(t), len(t[0])).zero_()]) \
                                            if len(t) != max_pairs else t for t in t_inp], 0))
            else:
                out_inputs.append(torch.stack(t_inp, 0))
        if transposed_targets is not None:
            for i, t_targ in enumerate(transposed_targets): #0:labels, 1:costs, 2:true_ants, 3:false_ants
                if i == 2 or i == 3:
                    if debug:
                        print("collate before", t_targ)
                    # shift the antecedent index associated to single anaphores (last)
                    t_targ = tuple(t.masked_fill_(torch.eq(t, len(t)-1), max_pairs) for t in t_targ)
                    if debug:
                        print("collate after", t_targ)
                out_targets.append(torch.stack(
                    [torch.cat(
                        [t[:-1] if len(t) > 2 else t.new(1).fill_(t[0]),
                        t.new(max_pairs + 1 - len(t)).fill_(t[0]),
                        t.new(1).fill_(t[-1])]
                    ) if len(t) != max_pairs + 1 else t for t in t_targ
                    ], 0))

            t_costs = transposed_targets[1] # We build the weights from the costs to have a float Tensor
            out_targets.append(torch.stack(
                    [torch.cat([t.new(len(t)-1).fill_(1),
                                t.new(max_pairs + 1 - len(t)).zero_(),
                                t.new(1).fill_(1)]) if len(t) != max_pairs + 1 \
                        else t.new(max_pairs + 1).fill_(1) for t in t_costs], 0))
        else:
            # Remark this mask is the inverse of the weights in the above target (used for evaluation masking)
            t_base = transposed_inputs[3]
            out_targets = torch.stack(
                    [torch.cat([t.new(len(t)-1).zero_().byte(),
                                t.new(max_pairs + 1 - len(t)).fill_(1).byte(),
                                t.new(1).zero_().byte()]) if len(t) != max_pairs + 1 \
                        else t.new(max_pairs + 1).zero_().byte() for t in t_base], 0)
    else:
        out_inputs = [torch.stack(t_inp, 0) for t_inp in transposed_inputs]
        if transposed_targets is not None:
            out_targets = [torch.stack(t_targ, 0) for t_targ in transposed_targets]
            out_targets.append(out_targets[1].new(len(out_targets[1]), 1).fill_(1))
        else:
            out_targets = out_inputs[0].new(len(out_inputs[0]), 1).zero_().byte()
    return (out_inputs, out_targets)



# coding: utf8
"""Conll training algorithm"""
from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

import os
import time
import argparse
import socket
from datetime import datetime
import numpy as np

import torch
import torch.nn as nn
from torch.autograd import Variable
from torch.optim import RMSprop
from torch.utils.data import DataLoader
from tensorboardX import SummaryWriter

from neuralcoref.model import Model
from neuralcoref.dataset import (NCDataset, NCBatchSampler,
    load_embeddings_from_file, padder_collate,
    SIZE_PAIR_IN, SIZE_SINGLE_IN, SIZE_EMBEDDING)
from neuralcoref.evaluator import ConllEvaluator

PACKAGE_DIRECTORY = os.path.dirname(os.path.abspath(__file__))
STAGES = ["allpairs", "toppairs", "ranking"]

def clipped_sigmoid(inputs):
    epsilon = 1.0e-7
    return torch.sigmoid(inputs).clamp(epsilon, 1.0 - epsilon)

def get_all_pairs_loss(n):
    def all_pair_loss(scores, targets):
        """ All pairs and single mentions probabilistic loss
        """
        labels = targets[0]
        weights = targets[4].data if len(targets) == 5 else None
        loss_op = nn.BCEWithLogitsLoss(weight=weights, size_average=False)
        loss = loss_op(scores, labels)
        return loss / n
    return all_pair_loss

def get_top_pair_loss(n):
    def top_pair_loss(scores, targets, debug=False):
        """ Top pairs (best true and best mistaken) and single mention probabilistic loss
        """
        true_ants = targets[2]
        false_ants = targets[3] if len(targets) == 5 else None
        s_scores = clipped_sigmoid(scores)
        true_pairs = torch.gather(s_scores, 1, true_ants)
        top_true, top_true_arg = torch.log(true_pairs).max(dim=1)  # max(log(p)), p=sigmoid(s)
        if debug:
            print("true_pairs", true_pairs.data)
            print("top_true", top_true.data)
            print("top_true_arg", top_true_arg.data)
        out_score = torch.sum(top_true).neg()
        if false_ants is not None: # We have no false antecedents when there are no pairs
            false_pairs = torch.gather(s_scores, 1, false_ants)
            top_false, _ = torch.log(1-false_pairs).min(dim=1) # min(log(1-p)), p=sigmoid(s)
            out_score = out_score + torch.sum(top_false).neg()
        return out_score / n
    return top_pair_loss

def get_ranking_loss(n):
    def ranking_loss(scores, targets):
        """ Slack-rescaled max margin loss
        """
        costs = targets[1]
        true_ants = targets[2]
        weights = targets[4] if len(targets) == 5 else None
        true_ant_score = torch.gather(scores, 1, true_ants)
        top_true, _ = true_ant_score.max(dim=1)
        tmp_loss = scores.add(1).add(top_true.unsqueeze(1).neg()) # 1 + scores - top_true
        if weights is not None:
            tmp_loss = tmp_loss.mul(weights)
        tmp_loss = tmp_loss.mul(costs)
        loss, _ = tmp_loss.max(dim=1)
        out_score = torch.sum(loss)
        return out_score / n
    return ranking_loss

def decrease_lr(optim_func, factor=0.1, min_lrs=0, eps=0, verbose=True):
    for i, param_group in enumerate(optim_func.param_groups):
        old_lr = float(param_group['lr'])
        new_lr = max(old_lr * factor, min_lrs)
        if old_lr - new_lr > eps:
            param_group['lr'] = new_lr
            if verbose:
                print('Reducing learning rate'
                      ' of group {} to {:.4e}.'.format(i, new_lr))
    return new_lr

def load_model(model, path):
    print("⛄️ Reloading model from", path)
    model.load_state_dict(torch.load(path) if args.cuda \
        else torch.load(path, map_location=lambda storage, loc: storage))

def run_model(args):
    print("Training for", args.all_pairs_epoch,
          args.top_pairs_epoch, args.ranking_epoch, "epochs")
    # Tensorboard server
    writer = SummaryWriter()

    # Load datasets and embeddings
    embed_path = args.weights if args.weights is not None else args.train
    tensor_embeddings, voc = load_embeddings_from_file(embed_path + "tuned_word")
    dataset = NCDataset(args.train, args)
    eval_dataset = NCDataset(args.eval, args)
    print("Vocabulary:", len(voc))

    # Construct model
    print("🏝 Build model")
    model = Model(len(voc), SIZE_EMBEDDING, args.h1,
                  args.h2, args.h3, SIZE_PAIR_IN, SIZE_SINGLE_IN)
    model.load_embeddings(tensor_embeddings)
    if args.cuda:
        model.cuda()
    if args.weights is not None:
        print("🏝 Loading pre-trained weights")
        model.load_weights(args.weights)
    if args.checkpoint_file is not None:
        print("⛄️ Loading model from", args.checkpoint_file)
        model.load_state_dict(torch.load(args.checkpoint_file) if args.cuda \
            else torch.load(args.checkpoint_file, map_location=lambda storage, loc: storage))

    print("🏝 Loading conll evaluator")
    eval_evaluator = ConllEvaluator(model, eval_dataset, args.eval, args.evalkey,
                               embed_path, args)
    train_evaluator = ConllEvaluator(model, dataset, args.train, args.trainkey,
                               embed_path, args)
    print("🏝 Testing evaluator and getting first eval score")
    eval_evaluator.test_model()
    start_time = time.time()
    eval_evaluator.build_test_file()
    score, f1_conll, ident = eval_evaluator.get_score()
    elapsed = time.time() - start_time
    print('|| s/evaluation {:5.2f}'.format(elapsed))
    writer.add_scalar("eval/" + "F1_conll", f1_conll, 0)

    # Preparing dataloader
    print("🏝 Preparing dataloader")
    print("Dataloader parameters: batchsize", args.batchsize, "numworkers", args.numworkers)
    batch_sampler = NCBatchSampler(dataset.mentions_pair_length, shuffle=True,
                                   batchsize=args.batchsize)
    dataloader = DataLoader(dataset, collate_fn=padder_collate, batch_sampler=batch_sampler,
                            num_workers=args.numworkers, pin_memory=args.cuda)
    mentions_idx, n_pairs = batch_sampler.get_batch_info()

    print("🏝 Start training")
    g_step = 0
    start_from = args.startstep if args.startstep is not None and args.startstage is not None else 0
    def run_epochs(start_epoch, end_epoch, loss_func, optim_func, save_name, lr, g_step, debug=None):
        best_model_path = args.save_path + "best_model" + save_name
        start_time_all = time.time()
        best_f1_conll = 0
        lower_eval = 0
        for epoch in range(start_epoch, end_epoch):
            """ Run an epoch """
            print("🚘 {} Epoch {:d}".format(save_name, epoch))
            model.train()
            start_time_log = time.time()
            start_time_epoch = time.time()
            epoch_loss = 0
            for batch_i, (m_idx, n_pairs_l, batch) in enumerate(zip(mentions_idx, n_pairs, dataloader)):
                if debug is not None and (debug == -1 or debug in m_idx):
                    l = list(dataset.flat_m_loc[m][2:] for m in m_idx)
                    print("🏔 Batch", batch_i, "m_idx:", "|".join(str(i) for i in m_idx),"mentions:","|".join(dataset.docs[d]['mentions'][i] for u, i, d in l))
                    print("Batch n_pairs:","|".join(str(p) for p in n_pairs_l))
                inputs, targets = batch
                inputs = tuple(Variable(inp, requires_grad=False) for inp in inputs)
                targets = tuple(Variable(tar, requires_grad=False) for tar in targets)
                if args.cuda:
                    inputs = tuple(i.cuda() for i in inputs)
                    targets = tuple(t.cuda() for t in targets)
                scores = model(inputs)
                if debug is not None and (debug == -1 or debug in m_idx):
                    print("Scores:\n" + "\n".join("|".join(str(s) for s in s_l) for s_l in scores.data.cpu().numpy()))
                    print("Labels:\n" + "\n".join("|".join(str(s) for s in s_l) for s_l in targets[0].data.cpu().numpy()))
                loss = loss_func(scores, targets)
                if debug is not None and (debug == -1 or debug in m_idx):
                    print('Loss', loss.data[0])
                # Zero gradients, perform a backward pass, and update the weights.
                optim_func.zero_grad()
                loss.backward()
                epoch_loss += loss.data[0]
                optim_func.step()
                writer.add_scalar("train/" + save_name + "_loss", loss.data[0], g_step)
                writer.add_scalar("meta/" + "lr", lr, g_step)
                writer.add_scalar("meta/" + "stage", STAGES.index(save_name), g_step)
                g_step += 1
                if batch_i % args.log_interval == 0 and batch_i > 0:
                    elapsed = time.time() - start_time_log
                    print('| epoch {:3d} | {:5d}/{:5d} batches | lr {:.2e} | ms/batch {:5.2f} | '
                            'loss {:.2e}'.format(
                        epoch, batch_i, len(dataloader), optim_func.param_groups[0]['lr'],
                        elapsed * 1000 / args.log_interval, loss.data[0]))
                    start_time_log = time.time()
            elapsed_all = time.time() - start_time_all
            elapsed_epoch = time.time() - start_time_epoch
            print('|| min/epoch {:5.2f} | est. remaining time (h) {:5.2f} | loss {:.2e}'.format(elapsed_epoch/60,
                elapsed_all/3600*float(end_epoch-epoch)/float(epoch-start_epoch+1), epoch_loss))
            writer.add_scalar("epoch/" + "loss", epoch_loss, g_step)
            if epoch % args.conll_train_interval == 0:
                start_time = time.time()
                train_evaluator.build_test_file()
                score, f1_conll, ident = train_evaluator.get_score()
                elapsed = time.time() - start_time
                print('|| min/train evaluation {:5.2f} | F1_conll {:5.2f}'.format(elapsed/60, f1_conll))
                writer.add_scalar("epoch/" + "F1_conll", f1_conll, g_step)
            if epoch % args.conll_eval_interval == 0:
                start_time = time.time()
                eval_evaluator.build_test_file()
                score, f1_conll, ident = eval_evaluator.get_score()
                elapsed = time.time() - start_time
                print('|| min/evaluation {:5.2f}'.format(elapsed/60))
                writer.add_scalar("eval/" + "F1_conll", f1_conll, g_step)
                g_step += 1
                save_path = args.save_path + save_name + "_" + str(epoch)
                torch.save(model.state_dict(), save_path)
                if f1_conll > best_f1_conll:
                    best_f1_conll = f1_conll
                    torch.save(model.state_dict(), best_model_path)
                    lower_eval = 0
                elif args.on_eval_decrease != 'nothing':
                    print("Evaluation metric decreases")
                    lower_eval += 1
                    if lower_eval >= args.patience:
                        if args.on_eval_decrease == 'divide_lr' or args.on_eval_decrease == 'divide_then_next':
                            print("reload best model and decrease lr")
                            load_model(model, best_model_path)
                            lr = decrease_lr(optim_func)
                        if args.on_eval_decrease == 'next_stage' or lr <= args.min_lr:
                            print("Switch to next stage")
                            break
        # Save last step
        start_time = time.time()
        eval_evaluator.build_test_file()
        score, f1_conll, ident = eval_evaluator.get_score()
        elapsed = time.time() - start_time
        print('|| min/evaluation {:5.2f}'.format(elapsed/60))
        writer.add_scalar("eval/" + "F1_conll", f1_conll, g_step)
        g_step += 1
        save_path = args.save_path + save_name + "_" + str(epoch)
        torch.save(model.state_dict(), save_path)
        load_model(model, best_model_path)
        return g_step

    if args.startstage is None or args.startstage == "allpairs":
        optimizer = RMSprop(model.parameters(), lr=args.all_pairs_lr, weight_decay=args.all_pairs_l2)
        loss_func = get_all_pairs_loss(batch_sampler.pairs_per_batch)
        g_step = run_epochs(start_from, args.all_pairs_epoch, loss_func, optimizer,
                            "allpairs", args.all_pairs_lr, g_step)
        start_from = 0

    if args.startstage is None or args.startstage in ["allpairs", "toppairs"]:
        optimizer = RMSprop(model.parameters(), lr=args.top_pairs_lr, weight_decay=args.top_pairs_l2)
        loss_func = get_top_pair_loss(10 * batch_sampler.mentions_per_batch)
        g_step = run_epochs(start_from, args.top_pairs_epoch, loss_func, optimizer,
                            "toppairs", args.top_pairs_lr, g_step)
        start_from = 0

    if args.startstage is None or args.startstage in ["ranking", "allpairs", "toppairs"]:
        optimizer = RMSprop(model.parameters(), lr=args.ranking_lr, weight_decay=args.ranking_l2)
        loss_func = get_ranking_loss(batch_sampler.mentions_per_batch)
        g_step = run_epochs(start_from, args.ranking_epoch, loss_func, optimizer,
                            "ranking", args.ranking_lr, g_step)

if __name__ == '__main__':
    DIR_PATH = os.path.dirname(os.path.realpath(__file__))
    parser = argparse.ArgumentParser(description='Training the neural coreference model')
    parser.add_argument('--train', type=str, default=DIR_PATH + '/data/tiny/', help='Path to the train dataset')
    parser.add_argument('--eval', type=str, default=DIR_PATH + '/data/tiny/', help='Path to the eval dataset')
    parser.add_argument('--evalkey', type=str, help='Path to an optional key file for scoring')
    parser.add_argument('--weights', type=str, help='Path to pre-trained weights (if you only want to test the scoring for e.g.)')
    parser.add_argument('--batchsize', type=int, default=20000, help='Size of a batch in total number of pairs')
    parser.add_argument('--numworkers', type=int, default=8, help='Number of workers for loading batches')
    parser.add_argument('--startstage', type=str, help='Start from a specific stage ("allpairs", "toppairs", "ranking")')
    parser.add_argument('--startstep', type=int, help='Start from a specific step')
    parser.add_argument('--checkpoint_file', type=str, help='Start from a previously saved checkpoint file')
    parser.add_argument('--log_interval', type=int, default=10, help='test every X mini-batches')
    parser.add_argument('--conll_eval_interval', type=int, default=10, help='evaluate eval F1 conll every X epochs')
    parser.add_argument('--conll_train_interval', type=int, default=20, help='evaluate train F1 conll every X epochs')
    parser.add_argument('--seed', type=int, default=1111, help='random seed')
    parser.add_argument('--costfn', type=float, default=0.8, help='cost of false new')
    parser.add_argument('--costfl', type=float, default=0.4, help='cost of false link')
    parser.add_argument('--costwl', type=float, default=1.0, help='cost of wrong link')
    parser.add_argument('--h1', type=int, default=1000, help='number of hidden unit on layer 1')
    parser.add_argument('--h2', type=int, default=500, help='number of hidden unit on layer 2')
    parser.add_argument('--h3', type=int, default=500, help='number of hidden unit on layer 3')
    parser.add_argument('--all_pairs_epoch', type=int, default=200, help='number of epochs for all-pairs pre-training')
    parser.add_argument('--top_pairs_epoch', type=int, default=200, help='number of epochs for top-pairs pre-training')
    parser.add_argument('--ranking_epoch', type=int, default=200, help='number of epochs for ranking training')
    parser.add_argument('--all_pairs_lr', type=float, default=2e-4, help='all pairs pre-training learning rate')
    parser.add_argument('--top_pairs_lr', type=float, default=2e-4, help='top pairs pre-training learning rate')
    parser.add_argument('--ranking_lr', type=float, default=2e-6, help='ranking training learning rate')
    parser.add_argument('--all_pairs_l2', type=float, default=1e-6, help='all pairs pre-training l2 regularization')
    parser.add_argument('--top_pairs_l2', type=float, default=1e-5, help='top pairs pre-training l2 regularization')
    parser.add_argument('--ranking_l2', type=float, default=1e-5, help='ranking training l2 regularization')
    parser.add_argument('--patience', type=int, default=3, help='patience (epochs) before considering evaluationhas decreased')
    parser.add_argument('--min_lr', type=float, default=2e-8, help='min learning rate')
    parser.add_argument('--on_eval_decrease', type=str, default='nothing',
                        help='What to do when evaluation decreases ("nothing", "divide_lr", "next_stage", "divide_then_next")')
    args = parser.parse_args()

    args.costs = {'FN': args.costfn, 'FL': args.costfl, 'WL' : args.costwl }

    current_time = datetime.now().strftime('%b%d_%H-%M-%S')
    args.save_path = os.path.join(PACKAGE_DIRECTORY, 'checkpoints', current_time + '_' + socket.gethostname() + '_')

    np.random.seed(args.seed)
    torch.manual_seed(args.seed)
    args.cuda = torch.cuda.is_available()
    if args.cuda:
        torch.cuda.manual_seed(args.seed)

    args.evalkey = args.evalkey if args.evalkey is not None else args.eval + "/key.txt"
    args.trainkey = args.train + "/key.txt"
    args.train = args.train + "/numpy/"
    args.eval = args.eval + "/numpy/"
    print(args)
    run_model(args)



# coding: utf8
"""Conll training algorithm"""
from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

import os
import numpy as np

import torch
import torch.nn as nn
import torch.utils.data

class Model(nn.Module):
    def __init__(self, vocab_size, embedding_dim, H1, H2, H3, D_pair_in, D_single_in, dropout=0.5):
        super(Model, self).__init__()
        self.word_embeds = nn.Embedding(vocab_size, embedding_dim)
        self.drop = nn.Dropout(dropout)
        self.pair_top = nn.Sequential(nn.Linear(D_pair_in, H1), nn.ReLU(), nn.Dropout(dropout),
                                      nn.Linear(H1, H2), nn.ReLU(), nn.Dropout(dropout),
                                      nn.Linear(H2, H3), nn.ReLU(), nn.Dropout(dropout),
                                      nn.Linear(H3, 1),
                                      nn.Linear(1, 1))
        self.single_top = nn.Sequential(nn.Linear(D_single_in, H1), nn.ReLU(), nn.Dropout(dropout),
                                        nn.Linear(H1, H2), nn.ReLU(), nn.Dropout(dropout),
                                        nn.Linear(H2, H3), nn.ReLU(), nn.Dropout(dropout),
                                        nn.Linear(H3, 1),
                                        nn.Linear(1, 1))
        self.init_weights()

    def init_weights(self):
        w = (param.data for name, param in self.named_parameters() if 'weight' in name)
        b = (param.data for name, param in self.named_parameters() if 'bias' in name)
        nn.init.uniform(self.word_embeds.weight.data, a=-0.5, b=0.5)
        for t in w:
            nn.init.xavier_uniform(t)
        for t in b:
            nn.init.constant(t, 0)

    def load_embeddings(self, preloaded_weights):
        self.word_embeds.weight = nn.Parameter(preloaded_weights)

    def load_weights(self, weights_path):
        print("Loading weights")
        single_layers_weights, single_layers_biases = [], []
        for f in sorted(os.listdir(weights_path)):
            if f.startswith("single_mention_weights"):
                single_layers_weights.append(np.load(os.path.join(weights_path, f)))
            if f.startswith("single_mention_bias"):
                single_layers_biases.append(np.load(os.path.join(weights_path, f)))
        top_single_linear = (layer for layer in self.single_top if isinstance(layer, nn.Linear))
        for w, b, layer in zip(single_layers_weights, single_layers_biases, top_single_linear):
            layer.weight = nn.Parameter(torch.from_numpy(w).float())
            layer.bias = nn.Parameter(torch.from_numpy(b).float().squeeze())
        pair_layers_weights, pair_layers_biases = [], []
        for f in sorted(os.listdir(weights_path)):
            if f.startswith("pair_mentions_weights"):
                pair_layers_weights.append(np.load(os.path.join(weights_path, f)))
            if f.startswith("pair_mentions_bias"):
                pair_layers_biases.append(np.load(os.path.join(weights_path, f)))
        top_pair_linear = (layer for layer in self.pair_top if isinstance(layer, nn.Linear))
        for w, b, layer in zip(pair_layers_weights, pair_layers_biases, top_pair_linear):
            layer.weight = nn.Parameter(torch.from_numpy(w).float())
            layer.bias = nn.Parameter(torch.from_numpy(b).float().squeeze())

    def forward(self, inputs, concat_axis=1):
        pairs = (len(inputs) == 8)
        if pairs:
            spans, words, single_features, ant_spans, ant_words, ana_spans, ana_words, pair_features = inputs
        else:
            spans, words, single_features = inputs
        embed_words = self.drop(self.word_embeds(words).view(words.size()[0], -1))
        single_input = torch.cat([spans, embed_words, single_features], 1)
        single_scores = self.single_top(single_input)
        if pairs:
            batchsize, pairs_num, _ = ana_spans.size()
            ant_embed_words = self.drop(self.word_embeds(ant_words.view(batchsize, -1)).view(batchsize, pairs_num, -1))
            ana_embed_words = self.drop(self.word_embeds(ana_words.view(batchsize, -1)).view(batchsize, pairs_num, -1))
            pair_input = torch.cat([ant_spans, ant_embed_words, ana_spans, ana_embed_words, pair_features], 2)
            pair_scores = self.pair_top(pair_input).squeeze(dim=2)
            total_scores = torch.cat([pair_scores, single_scores], concat_axis)
        return total_scores if pairs else single_scores



# coding: utf8
"""data models and pre-processing for the coref algorithm"""

from __future__ import unicode_literals
from __future__ import print_function

import re
import io
from six import string_types, integer_types

from neuralcoref.compat import unicode_
from neuralcoref.utils import encode_distance, parallel_process

try:
    from itertools import izip_longest as zip_longest
except ImportError: # will be 3.x series
    from itertools import zip_longest

import spacy
import numpy as np

#########################
####### UTILITIES #######
#########################

NO_COREF_LIST = ["i", "me", "my", "you", "your"]

MENTION_TYPE = {"PRONOMINAL": 0, "NOMINAL": 1, "PROPER": 2, "LIST": 3}
MENTION_LABEL = {0: "PRONOMINAL", 1: "NOMINAL", 2: "PROPER", 3: "LIST"}

PROPERS_TAGS = ["NN", "NNS", "NNP", "NNPS"]
ACCEPTED_ENTS = ["PERSON", "NORP", "FACILITY", "ORG", "GPE", "LOC", "PRODUCT", "EVENT", "WORK_OF_ART", "LANGUAGE"]
WHITESPACE_PATTERN = r"\s+|_+"
UNKNOWN_WORD = "*UNK*"
MISSING_WORD = "<missing>"
MAX_ITER = 100

#########################
## MENTION EXTRACTION ###
#########################

def extract_mentions_spans(doc, blacklist=True, debug=False):
    '''
    Extract potential mentions from a spacy parsed Doc
    '''
    if debug: print('===== doc ====:', doc)
    for c in doc:
        if debug: print("🚧 span search:", c, "head:", c.head, "tag:", c.tag_, "pos:", c.pos_, "dep:", c.dep_)
    # Named entities
    mentions_spans = list(ent for ent in doc.ents if ent.label_ in ACCEPTED_ENTS)

    if debug: print("==-- ents:", list(((ent, ent.label_) for ent in mentions_spans)))
    for spans in parallel_process([{'doc': doc,
                                    'span': sent,
                                    'blacklist': blacklist} for sent in doc.sents],
                                _extract_from_sent, use_kwargs=True, front_num=0):
        mentions_spans = mentions_spans + spans
    spans_set = set()
    cleaned_mentions_spans = []
    for spans in mentions_spans:
        if spans.end > spans.start and (spans.start, spans.end) not in spans_set:
            cleaned_mentions_spans.append(spans)
            spans_set.add((spans.start, spans.end))

    return cleaned_mentions_spans

def _extract_from_sent(doc, span, blacklist=True, debug=False):
    '''
    Extract Pronouns and Noun phrases mentions from a spacy Span
    '''
    keep_tags = re.compile(r"N.*|PRP.*|DT|IN")
    leave_dep = ["det", "compound", "appos"]
    keep_dep = ["nsubj", "dobj", "iobj", "pobj"]
    nsubj_or_dep = ["nsubj", "dep"]
    conj_or_prep = ["conj", "prep"]
    remove_pos = ["CCONJ", "INTJ", "ADP"]
    lower_not_end = ["'s", ',', '.', '!', '?', ':', ';']

    # Utility to remove bad endings
    def cleanup_endings(left, right, token):
        minchild_idx = min(left + [token.i]) if left else token.i
        maxchild_idx = max(right + [token.i]) if right else token.i
        # Clean up endings and begginging
        while maxchild_idx >= minchild_idx and (doc[maxchild_idx].pos_ in remove_pos
                                           or doc[maxchild_idx].lower_ in lower_not_end):
            if debug: print("Removing last token", doc[maxchild_idx].lower_, doc[maxchild_idx].tag_)
            maxchild_idx -= 1 # We don't want mentions finishing with 's or conjunctions/punctuation
        while minchild_idx <= maxchild_idx and (doc[minchild_idx].pos_ in remove_pos 
                                           or doc[minchild_idx].lower_ in lower_not_end):
            if debug: print("Removing first token", doc[minchild_idx].lower_, doc[minchild_idx].tag_)
            minchild_idx += 1 # We don't want mentions starting with 's or conjunctions/punctuation
        return minchild_idx, maxchild_idx+1

    mentions_spans = []
    for token in span:
        if debug: print("🚀 tok:", token, "tok.tag_:", token.tag_, "tok.pos_:", token.pos_, "tok.dep_:", token.dep_)

        if blacklist and token.lower_ in NO_COREF_LIST:
            if debug: print("token in no_coref_list")
            continue
        if (not keep_tags.match(token.tag_) or token.dep_ in leave_dep) and not token.dep_ in keep_dep:
            if debug: print("not pronoun or no right dependency")
            continue

        # pronoun
        if re.match(r"PRP.*", token.tag_):
            if debug: print("PRP")
            endIdx = token.i + 1

            span = doc[token.i: endIdx]
            if debug: print("==-- PRP store:", span)
            mentions_spans.append(span)

            # when pronoun is a part of conjunction (e.g., you and I)
            if token.n_rights > 0 or token.n_lefts > 0:
                span = doc[token.left_edge.i : token.right_edge.i+1]
                if debug: print("==-- in conj store:", span)
                mentions_spans.append(span)
            continue

        # Add NP mention
        if debug:
            print("NP or IN:", token.lower_)
            if token.tag_ == 'IN':
                print("IN tag")
        # Take care of 's
        if token.lower_ == "'s":
            if debug: print("'s detected")
            h = token.head
            j = 0
            while h.head.i != h.i and j < MAX_ITER:
                if debug:
                    print("token head:", h, h.dep_, "head:", h.head)
                    print(id(h.head), id(h))
                if h.dep_ == "nsubj":
                    minchild_idx = min((c.left_edge.i for c in doc if c.head.i == h.head.i and c.dep_ in nsubj_or_dep),
                                       default=token.i)
                    maxchild_idx = max((c.right_edge.i for c in doc if c.head.i == h.head.i and c.dep_ in nsubj_or_dep),
                                       default=token.i)
                    if debug: print("'s', i1:", doc[minchild_idx], " i2:", doc[maxchild_idx])
                    span = doc[minchild_idx : maxchild_idx+1]
                    if debug: print("==-- 's' store:", span)
                    mentions_spans.append(span)
                    break
                h = h.head
                j += 1
            assert j != MAX_ITER
            continue

        # clean up
        for c in doc:
            if debug and c.head.i == token.i: print("🚧 token in span:", c, "- head & dep:", c.head, c.dep_)
        left = list(c.left_edge.i for c in doc if c.head.i == token.i)
        right = list(c.right_edge.i for c in doc if c.head.i == token.i)
        if token.tag_ == 'IN' and token.dep_ == "mark" and len(left) == 0 and len(right) == 0:
            left = list(c.left_edge.i for c in doc if c.head.i == token.head.i)
            right = list(c.right_edge.i for c in doc if c.head.i == token.head.i)
        if debug:
            print("left side:", left)
            print("right side:", right)
            minchild_idx = min(left) if left else token.i
            maxchild_idx = max(right) if right else token.i
            print("full span:", doc[minchild_idx:maxchild_idx+1])
        start, end = cleanup_endings(left, right, token)
        if start == end:
            continue
        if doc[start].lower_ == "'s":
            continue # we probably already have stored this mention
        span = doc[start:end]
        if debug:
            print("cleaned endings span:", doc[start:end])
            print("==-- full span store:", span)
        mentions_spans.append(span)
        if debug and token.tag_ == 'IN':
            print("IN tag")
        if any(tok.dep_ in conj_or_prep for tok in span):
            if debug: print("Conjunction found, storing first element separately")
            for c in doc:
                if c.head.i == token.i and c.dep_ not in conj_or_prep:
                    if debug: print("left no conj:", c, 'dep & edge:', c.dep_, c.left_edge)
                    if debug: print("right no conj:", c, 'dep & edge:', c.dep_, c.right_edge)
            left_no_conj = list(c.left_edge.i for c in doc if c.head.i == token.i and c.dep_ not in conj_or_prep)
            right_no_conj = list(c.right_edge.i for c in doc if c.head.i == token.i and c.dep_ not in conj_or_prep)
            if debug: print("left side no conj:", [doc[i] for i in left_no_conj])
            if debug: print("right side no conj:", [doc[i] for i in right_no_conj])
            start, end = cleanup_endings(left_no_conj, right_no_conj, token)
            if start == end:
                continue
            span = doc[start:end]
            if debug: print("==-- full span store:", span)
            mentions_spans.append(span)
    if debug: print("mentions_spans inside", mentions_spans)
    return mentions_spans

#########################
####### CLASSES #########

class Mention(spacy.tokens.Span):
    '''
    A mention (possible anaphor) inherite from spacy Span class with additional informations
    '''
    def __new__(cls, span, mention_index, utterance_index, utterance_start_sent, speaker=None, gold_label=None, *args, **kwargs):
        # We need to override __new__ see http://cython.readthedocs.io/en/latest/src/userguide/special_methods.html
        obj = spacy.tokens.Span.__new__(cls, span.doc, span.start, span.end, *args, **kwargs)
        return obj

    def __init__(self, span, mention_index, utterance_index, utterances_start_sent,
                 speaker=None, gold_label=None):
        '''
        Arguments:
            span (spaCy Span): the spaCy span from which creating the Mention object
            mention_index (int): index of the Mention in the Document
            utterance_index (int): index of the utterance of the Mention in the Document
            utterances_start_sent (int): index of the first sentence of the utterance of the Mention in the Document
                (an utterance can comprise several sentences)
            speaker (Speaker): the speaker of the mention
            gold_label (anything): a gold label associated to the Mention (for training)
        '''
        self.index = mention_index
        self.utterance_index = utterance_index
        self.utterances_sent = utterances_start_sent + self._get_doc_sent_number()
        self.speaker = speaker
        self.gold_label = gold_label
        self.spans_embeddings = None
        self.words_embeddings = None
        self.features = None

        self.features_ = None
        self.spans_embeddings_ = None
        self.words_embeddings_ = None

        self.mention_type = self._get_type()
        self.propers = set(self.content_words)
        self.entity_label = self._get_entity_label()
        self.in_entities = self._get_in_entities()

    def _get_entity_label(self):
        ''' Label of a detected named entity the Mention is nested in if any'''
        for ent in self.doc.ents:
            if ent.start <= self.start and self.end <= ent.end:
                return ent.label
        return None

    def _get_in_entities(self):
        ''' Is the Mention nested in a detected named entity'''
        return self.entity_label is not None

    def _get_type(self):
        ''' Find the type of the Span '''
        conj = ["CC", ","]
        prp = ["PRP", "PRP$"]
        proper = ["NNP", "NNPS"]
        if any(t.tag_ in conj and t.ent_type_ not in ACCEPTED_ENTS for t in self):
            mention_type = MENTION_TYPE["LIST"]
        elif self.root.tag_ in prp:
            mention_type = MENTION_TYPE["PRONOMINAL"]
        elif self.root.ent_type_ in ACCEPTED_ENTS or self.root.tag_ in proper:
            mention_type = MENTION_TYPE["PROPER"]
        else:
            mention_type = MENTION_TYPE["NOMINAL"]
        return mention_type

    def _get_doc_sent_number(self):
        ''' Index of the sentence of the Mention in the current utterance'''
        for i, s in enumerate(self.doc.sents):
            if s == self.sent:
                return i
        return None

    @property
    def content_words(self):
        ''' Returns an iterator of nouns/proper nouns in the Mention '''
        return (tok.lower_ for tok in self if tok.tag_ in PROPERS_TAGS)

    @property
    def embedding(self):
        return np.concatenate([self.spans_embeddings, self.words_embeddings], axis=0)

    def heads_agree(self, mention2):
        ''' Does the root of the Mention match the root of another Mention/Span'''
        # we allow same-type NEs to not match perfectly,
        # but rather one could be included in the other, e.g., "George" -> "George Bush"
        if (self.in_entities and mention2.in_entities and
                self.entity_label == mention2.entity_label and
                (self.root.lower_ in mention2.lower_ or mention2.root.lower_ in self.lower_)):
            return True
        return self.root.lower_ == mention2.root.lower_

    def exact_match(self, mention2):
        ''' Does the Mention lowercase text matches another Mention/Span lowercase text'''
        return self.lower_ == mention2.lower_

    def relaxed_match(self, mention2):
        ''' Does the nouns/proper nous in the Mention match another Mention/Span nouns/propers'''
        return not self.propers.isdisjoint(mention2.propers)

    def speaker_match_mention(self, mention2):
        # To take care of sentences like 'Larry said, "San Francisco is a city."': (id(Larry), id(San Francisco))
        # if document.speakerPairs.contains(new Pair<>(mention.mentionID, ant.mentionID)):
        #    return True
        if self.speaker is not None:
            return self.speaker.speaker_matches_mention(mention2, strict_match=False)
        return False

class Speaker(object):
    '''
    A speaker with its names, list of mentions and matching test functions
    '''
    def __init__(self, speaker_id, speaker_names=None):
        self.mentions = []
        self.speaker_id = speaker_id
        if speaker_names is None:
            self.speaker_names = [unicode_(speaker_id)]
        elif isinstance(speaker_names, string_types):
            self.speaker_names = [speaker_names]
        elif len(speaker_names) > 1:
            self.speaker_names = speaker_names
        else:
            self.speaker_names = unicode_(speaker_names)
        self.speaker_tokens = [tok.lower() for s in self.speaker_names for tok in re.split(WHITESPACE_PATTERN, s)]

    def __str__(self):
        return '{} <names> {}'.format(self.speaker_id, self.speaker_names)

    def add_mention(self, mention):
        ''' Add a Mention of the Speaker'''
        self.mentions.append(mention)

    def contain_mention(self, mention):
        ''' Does the Speaker contains a Mention'''
        return mention in self.mentions

    def contain_string(self, string):
        ''' Does the Speaker names contains a string'''
        return any(re.sub(WHITESPACE_PATTERN, "", string.lower()) == re.sub(WHITESPACE_PATTERN, "", s.lower())
                   for s in self.speaker_names)

    def contain_token(self, token):
        ''' Does the Speaker names contains a token (word)'''
        return any(token.lower() == tok.lower() for tok in self.speaker_tokens)

    def speaker_matches_mention(self, mention, strict_match=False):
        ''' Does a Mention matches the speaker names'''
         # Got info about this speaker
        if self.contain_mention(mention):
            return True
        if strict_match:
            if self.contain_string(mention):
                self.mentions.append(mention)
                return True
        else:
            # test each token in the speaker names
            if not mention.root.tag_.startswith("NNP"):
                return False
            if self.contain_token(mention.root.lower_):
                self.mentions.append(mention)
                return True
        return False

class EmbeddingExtractor:
    '''
    Compute words embedding features for mentions
    '''
    def __init__(self, pretrained_model_path):
        _, self.static_embeddings, self.stat_idx, self.stat_voc = self.load_embeddings_from_file(pretrained_model_path + "static_word")
        _, self.tuned_embeddings, self.tun_idx, self.tun_voc = self.load_embeddings_from_file(pretrained_model_path + "tuned_word")
        self.fallback = self.static_embeddings.get(UNKNOWN_WORD)

        self.shape = self.static_embeddings[UNKNOWN_WORD].shape
        shape2 = self.tuned_embeddings[UNKNOWN_WORD].shape
        assert self.shape == shape2

    @staticmethod
    def load_embeddings_from_file(name):
        print("Loading embeddings from", name)
        embeddings = {}
        voc_to_idx = {}
        idx_to_voc = []
        mat = np.load(name+"_embeddings.npy")
        average_mean = np.average(mat, axis=0, weights=np.sum(mat, axis=1))
        with io.open(name+"_vocabulary.txt", 'r', encoding='utf-8') as f:
            for i, line in enumerate(f):
                embeddings[line.strip()] = mat[i, :]
                voc_to_idx[line.strip()] = i
                idx_to_voc.append(line.strip())
        return average_mean, embeddings, voc_to_idx, idx_to_voc

    @staticmethod
    def normalize_word(w):
        if w is None:
            return MISSING_WORD
        return re.sub(r"\d", u"0", w.lower_)

    def get_document_embedding(self, utterances_list):
        ''' Embedding for the document '''
    #    We could also use this: embed_vector = np.copy(self.average_mean)#np.zeros(self.shape)
    #    return embed_vector
        embed_vector = np.zeros(self.shape)
        for utt in utterances_list:
            _, utt_embed = self.get_average_embedding(utt)
            embed_vector += utt_embed
        return embed_vector/max(len(utterances_list), 1)

    def get_stat_word(self, word):
        if word in self.static_embeddings:
            return word, self.static_embeddings.get(word)
        else:
            return UNKNOWN_WORD, self.fallback

    def get_word_embedding(self, word, static=False):
        ''' Embedding for a single word (tuned if possible, otherwise static) '''
        norm_word = self.normalize_word(word)
        if static:
            return self.get_stat_word(norm_word)
        else:
            if norm_word in self.tuned_embeddings:
                return norm_word, self.tuned_embeddings.get(norm_word)
            else:
                return self.get_stat_word(norm_word)
 
    def get_word_in_sentence(self, word_idx, sentence):
        ''' Embedding for a word in a sentence '''
        if word_idx < sentence.start or word_idx >= sentence.end:
            return self.get_word_embedding(None)
        return self.get_word_embedding(sentence.doc[word_idx])

    def get_average_embedding(self, token_list):
        ''' Embedding for a list of words '''
        embed_vector = np.zeros(self.shape) #We could also use np.copy(self.average_mean)
        word_list = []
        for tok in token_list:
            if tok.lower_ not in [".", "!", "?"]:
                word, embed = self.get_word_embedding(tok, static=True)
                embed_vector += embed
                word_list.append(word)
        return word_list, (embed_vector/max(len(word_list), 1))

    def get_mention_embeddings(self, mention, doc_embedding):
        ''' Get span (averaged) and word (single) embeddings of a mention '''
        st = mention.sent
        mention_lefts = mention.doc[max(mention.start-5, st.start):mention.start]
        mention_rights = mention.doc[mention.end:min(mention.end+5, st.end)]
        head = mention.root.head
        spans = [self.get_average_embedding(mention),
                 self.get_average_embedding(mention_lefts),
                 self.get_average_embedding(mention_rights),
                 self.get_average_embedding(st),
                 (unicode_(doc_embedding[0:8]) + "...", doc_embedding)]
        words = [self.get_word_embedding(mention.root),
                 self.get_word_embedding(mention[0]),
                 self.get_word_embedding(mention[-1]),
                 self.get_word_in_sentence(mention.start-1, st),
                 self.get_word_in_sentence(mention.end, st),
                 self.get_word_in_sentence(mention.start-2, st),
                 self.get_word_in_sentence(mention.end+1, st),
                 self.get_word_embedding(head)]
        spans_embeddings_ = {"00_Mention": spans[0][0],
                             "01_MentionLeft": spans[1][0],
                             "02_MentionRight": spans[2][0],
                             "03_Sentence": spans[3][0],
                             "04_Doc": spans[4][0]}
        words_embeddings_ = {"00_MentionHead": words[0][0],
                             "01_MentionFirstWord": words[1][0],
                             "02_MentionLastWord": words[2][0],
                             "03_PreviousWord": words[3][0],
                             "04_NextWord": words[4][0],
                             "05_SecondPreviousWord": words[5][0],
                             "06_SecondNextWord": words[6][0],
                             "07_MentionRootHead": words[7][0]}
        return (spans_embeddings_,
                words_embeddings_,
                np.concatenate([em[1] for em in spans], axis=0),
                np.concatenate([em[1] for em in words], axis=0))

class Document(object):
    '''
    Main data class: encapsulate list of utterances, mentions and speakers
    Process utterances to extract mentions and pre-compute mentions features
    '''
    def __init__(self, nlp, utterances=None, utterances_speaker=None, speakers_names=None,
                 blacklist=False, consider_speakers=False,
                 model_path=None, embedding_extractor=None,
                 conll=None, debug=False):
        '''
        Arguments:
            nlp (spaCy Language Class): A spaCy Language Class for processing the text input
            utterances: utterance(s) to load already see self.add_utterances()
            utterances_speaker: speaker(s) of utterance(s) to load already see self.add_utterances()
            speakers_names: speaker(s) of utterance(s) to load already see self.add_utterances()
            blacklist (boolean): use a list of term for which coreference is not preformed
            consider_speakers (boolean): consider speakers informations
            pretrained_model_path (string): Path to a folder with pretrained word embeddings
            embedding_extractor (EmbeddingExtractor): Use a pre-loaded word embeddings extractor
            conll (string): If training on coNLL data: identifier of the document type
            debug (boolean): print debug informations
        '''
        self.nlp = nlp
        self.blacklist = blacklist
        self.utterances = []
        self.utterances_speaker = []
        self.last_utterances_loaded = []
        self.mentions = []
        self.speakers = {}
        self.n_sents = 0
        self.debug = debug
        self.consider_speakers = consider_speakers or conll is not None

        self.genre_, self.genre = self.set_genre(conll)

        if model_path is not None and embedding_extractor is None:
            self.embed_extractor = EmbeddingExtractor(model_path)
        elif embedding_extractor is not None:
            self.embed_extractor = embedding_extractor
        else:
            self.embed_extractor = None

        if utterances:
            self.add_utterances(utterances, utterances_speaker, speakers_names)

    def set_genre(self, conll):
        if conll is not None:
            genre = np.zeros((7,))
            genre[conll] = 1
        else:
            genre = np.array(0, ndmin=1, copy=False)
        return conll, genre

    def __str__(self):
        return '<utterances, speakers> \n {}\n<mentions> \n {}' \
                .format('\n '.join(unicode_(i) + " " + unicode_(s) for i, s in zip(self.utterances, self.utterances_speaker)),
                        '\n '.join(unicode_(i) + " " + unicode_(i.speaker) for i in self.mentions))

    def __len__(self):
        ''' Return the number of mentions (not utterances) since it is what we really care about '''
        return len(self.mentions)

    def __getitem__(self, key):
        ''' Return a specific mention (not utterance) '''
        return self.mentions[key]

    def __iter__(self):
        ''' Iterate over mentions (not utterances) '''
        for mention in self.mentions:
            yield mention

    #######################################
    ###### UTERANCE LOADING FUNCTIONS #####
    #######################################
    
    def set_utterances(self, utterances, utterances_speaker=None, speakers_names=None):
        self.utterances = []
        self.utterances_speaker = []
        self.mentions = []
        self.speakers = {}
        self.n_sents = 0
        if utterances:
            self.add_utterances(utterances, utterances_speaker, speakers_names)

    def add_utterances(self, utterances, utterances_speaker=None, speakers_names=None):
        '''
        Add utterances to the utterance list and build mention list for these utterances

        Arg:
            utterances : iterator or list of string corresponding to successive utterances
            utterances_speaker : iterator or list of speaker id for each utterance.
                If not provided, assume two speakers speaking alternatively.
                if utterances and utterances_speaker are not of the same length padded with None
            speakers_names : dictionnary of list of acceptable speaker names for each speaker id
        Return:
            List of indexes of added utterances in the docs
        '''
        if self.debug: print("Adding utterances", utterances)
        if isinstance(utterances, string_types):
            utterances = [utterances]
        if utterances_speaker is None:
            if self.debug: print("No utterance speaker indication")
            a = -1
            if self.utterances_speaker and isinstance(self.utterances_speaker[-1].speaker_id, integer_types):
                a = self.utterances_speaker[-1].speaker_id
            utterances_speaker = ((i + a + 1) % 2 for i in range(len(utterances)))
        utterances_index = []
        utt_start = len(self.utterances)
        docs = list(self.nlp.pipe(utterances))
        m_spans = list(extract_mentions_spans(doc, blacklist=self.blacklist) for doc in docs)
        for utt_index, (doc, m_spans, speaker_id) in enumerate(zip_longest(docs, m_spans, utterances_speaker)):
            if speaker_id not in self.speakers:
                speaker_name = speakers_names.get(speaker_id, None) if speakers_names else None
                self.speakers[speaker_id] = Speaker(speaker_id, speaker_name)
            self._process_mentions(m_spans, utt_start + utt_index, self.n_sents, self.speakers[speaker_id])
            utterances_index.append(utt_start + utt_index)
            self.utterances.append(doc)
            self.utterances_speaker.append(self.speakers[speaker_id])
            self.n_sents += len(list(doc.sents))

        self.set_mentions_features()
        self.last_utterances_loaded = utterances_index

    ###################################
    ## FEATURES MENTIONS EXTRACTION ###
    ###################################

    def _process_mentions(self, mentions_spans, utterance_index, n_sents, speaker):
        '''
        Process mentions in a spacy doc (an utterance)
        '''
        processed_spans = sorted((m for m in mentions_spans), key=lambda m: (m.root.i, m.start))
        n_mentions = len(self.mentions)
        for mention_index, span in enumerate(processed_spans):
            self.mentions.append(Mention(span, mention_index + n_mentions,
                                             utterance_index, n_sents, speaker))

    def set_mentions_features(self):
        '''
        Compute features for the extracted mentions
        '''
        doc_embedding = self.embed_extractor.get_document_embedding(self.utterances) if self.embed_extractor is not None else None
        for mention in self.mentions:
            one_hot_type = np.zeros((4,))
            one_hot_type[mention.mention_type] = 1
            features_ = {"01_MentionType": mention.mention_type,
                         "02_MentionLength": len(mention)-1,
                         "03_MentionNormLocation": (mention.index)/len(self.mentions),
                         "04_IsMentionNested": 1 if any((m is not mention
                                                          and m.utterances_sent == mention.utterances_sent
                                                          and m.start <= mention.start
                                                          and mention.end <= m.end)
                                                         for m in self.mentions) else 0}
            features = np.concatenate([one_hot_type,
                                       encode_distance(features_["02_MentionLength"]),
                                       np.array(features_["03_MentionNormLocation"], ndmin=1, copy=False),
                                       np.array(features_["04_IsMentionNested"], ndmin=1, copy=False)
                                      ], axis=0)
            (spans_embeddings_, words_embeddings_,
             spans_embeddings, words_embeddings) = self.embed_extractor.get_mention_embeddings(mention, doc_embedding)
            mention.features_ = features_
            mention.features = features
            mention.spans_embeddings = spans_embeddings
            mention.spans_embeddings_ = spans_embeddings_
            mention.words_embeddings = words_embeddings
            mention.words_embeddings_ = words_embeddings_

    def get_single_mention_features(self, mention):
        ''' Features for anaphoricity test (signle mention features + genre if conll)'''
        features_ = mention.features_
        features_["DocGenre"] = self.genre_
        return (features_, np.concatenate([mention.features, self.genre], axis=0))

    def get_pair_mentions_features(self, m1, m2):
        ''' Features for pair of mentions (same speakers, speaker mentioned, string match)'''
        features_ = {"00_SameSpeaker": 1 if self.consider_speakers and m1.speaker == m2.speaker else 0,
                     "01_AntMatchMentionSpeaker": 1 if self.consider_speakers and m2.speaker_match_mention(m1) else 0,
                     "02_MentionMatchSpeaker": 1 if self.consider_speakers and m1.speaker_match_mention(m2) else 0,
                     "03_HeadsAgree": 1 if m1.heads_agree(m2) else 0,
                     "04_ExactStringMatch": 1 if m1.exact_match(m2) else 0,
                     "05_RelaxedStringMatch": 1 if m1.relaxed_match(m2) else 0,
                     "06_SentenceDistance": m2.utterances_sent - m1.utterances_sent,
                     "07_MentionDistance": m2.index - m1.index - 1,
                     "08_Overlapping": 1 if (m1.utterances_sent == m2.utterances_sent and m1.end > m2.start) else 0,
                     "09_M1Features": m1.features_,
                     "10_M2Features": m2.features_,
                     "11_DocGenre": self.genre_}
        pairwise_features = [np.array([features_["00_SameSpeaker"],
                                       features_["01_AntMatchMentionSpeaker"],
                                       features_["02_MentionMatchSpeaker"],
                                       features_["03_HeadsAgree"],
                                       features_["04_ExactStringMatch"],
                                       features_["05_RelaxedStringMatch"]]),
                             encode_distance(features_["06_SentenceDistance"]),
                             encode_distance(features_["07_MentionDistance"]),
                             np.array(features_["08_Overlapping"], ndmin=1),
                             m1.features,
                             m2.features,
                             self.genre]
        return (features_, np.concatenate(pairwise_features, axis=0))

    ###################################
    ###### ITERATOR OVER MENTIONS #####
    ###################################

    def get_candidate_mentions(self, last_utterances_added=False):
        '''
        Return iterator over indexes of mentions in a list of utterances if specified
        '''
        if last_utterances_added:
            for i, mention in enumerate(self.mentions):
                if self.debug: print("🤣", i, mention, "utterance index", mention.utterance_index)
                if mention.utterance_index in self.last_utterances_loaded:
                    yield i
        else:
            iterator = range(len(self.mentions))
            for i in iterator:
                yield i

    def get_candidate_pairs(self, mentions=None, max_distance=50, max_distance_with_match=500, debug=False):
        '''
        Yield tuples of mentions, dictionnary of candidate antecedents for the mention

        Arg:
            mentions: an iterator over mention indexes (as returned by get_candidate_mentions)
            max_mention_distance : max distance between a mention and its antecedent
            max_mention_distance_string_match : max distance between a mention and
                its antecedent when there is a proper noun match
        '''
        if mentions is None:
            mentions = range(len(self.mentions))
        if debug: print("get_candidate_pairs: mentions", mentions)
        
        if max_distance_with_match is not None:
            word_to_mentions = {}
            for i in mentions:
                for tok in self.mentions[i].content_words:
                    if not tok in word_to_mentions:
                        word_to_mentions[tok] = [i]
                    else:
                        word_to_mentions[tok].append(i)

        for i in mentions:
            antecedents = set(range(i)) if max_distance is None else set(range(max(0, i - max_distance), i))
            if debug: print("antecedents", antecedents)
            if max_distance_with_match is not None:
                for tok in self.mentions[i].content_words:
                    with_string_match = word_to_mentions.get(tok, None)
                    for match_idx in with_string_match:
                        if match_idx < i and match_idx >= i - max_distance_with_match:
                            antecedents.add(match_idx)
            yield i, antecedents

def mention_detection_debug(sentence):
    print(u"🌋 Loading spacy model")
    try:
        spacy.info('en_core_web_sm')
        model = 'en_core_web_sm'
    except IOError:
        print("No spacy 2 model detected, using spacy1 'en' model")
        spacy.info('en')
        model = 'en'
    nlp = spacy.load(model)
    doc = nlp(sentence.decode('utf-8'))
    mentions = extract_mentions_spans(doc, blacklist=False, debug=True)
    for mention in mentions:
        print(mention)

if __name__ == '__main__':
    import sys
    if len(sys.argv) > 1:
        sent = sys.argv[1]
        mention_detection_debug(sent)
    else:
        mention_detection_debug(u"My sister has a dog. She loves him.")



# coding: utf8
"""Conll parser"""
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

import re
import sys
import codecs
import argparse
import time
import os
import io
import pickle

import spacy
from spacy.tokens import Doc

import numpy as np

from tqdm import tqdm

from .compat import unicode_
from .document import Mention, Document, Speaker, EmbeddingExtractor, MISSING_WORD
from .utils import parallel_process

PACKAGE_DIRECTORY = os.path.dirname(os.path.abspath(__file__))
REMOVED_CHAR = ["/", "%", "*"]
NORMALIZE_DICT = {"/.": ".",
                  "/?": "?",
                  "-LRB-": "(",
                  "-RRB-": ")",
                  "-LCB-": "{",
                  "-RCB-": "}",
                  "-LSB-": "[",
                  "-RSB-": "]"}

CONLL_GENRES = {"bc": 0, "bn": 1, "mz": 2, "nw": 3, "pt": 4, "tc": 5, "wb": 6}

FEATURES_NAMES = ["mentions_features",          # 0
                  "mentions_labels",            # 1
                  "mentions_pairs_length",      # 2
                  "mentions_pairs_start_index", # 3
                  "mentions_spans",             # 4
                  "mentions_words",             # 5
                  "pairs_ant_index",            # 6
                  "pairs_features",             # 7
                  "pairs_labels",               # 8
                  "locations",                  # 9
                  "conll_tokens",               # 10
                  "spacy_lookup",               # 11
                  "doc",                        # 12
                  ]

MISSED_MENTIONS_FILE = os.path.join(PACKAGE_DIRECTORY, "test_mentions_identification.txt")
SENTENCES_PATH = os.path.join(PACKAGE_DIRECTORY, "test_sentences.txt")

###################
### UTILITIES #####

def clean_token(token):
    cleaned_token = token
    if cleaned_token in NORMALIZE_DICT:
        cleaned_token = NORMALIZE_DICT[cleaned_token]
    if cleaned_token not in REMOVED_CHAR:
        for char in REMOVED_CHAR:
            cleaned_token = cleaned_token.replace(char, u'')
    if len(cleaned_token) == 0:
        cleaned_token = ","
    return cleaned_token

def mention_words_idx(embed_extractor, mention, debug=False):
    # index of the word in the tuned embeddings no need for normalizing,
    # it is already performed in set_mentions_features()
    # We take them in the tuned vocabulary which is a smaller voc tailored from conll
    words = []
    for _, w in sorted(mention.words_embeddings_.items()):
        if w not in embed_extractor.tun_idx:
            if debug: print("No matching tokens in tuned voc for word ", w, "surrounding or inside mention", mention)
            words.append(MISSING_WORD)
        else:
            words.append(w)
    return [embed_extractor.tun_idx[w] for w in words]

def check_numpy_array(feature, array, n_mentions_list, compressed=True):
    for n_mentions in n_mentions_list:
        if feature == FEATURES_NAMES[0]:
            assert array.shape[0] == len(n_mentions)
            if compressed:
                assert np.array_equiv(array[:, 3], np.array([len(n_mentions)] * len(n_mentions)))
                assert np.max(array[:, 2]) == len(n_mentions)-1
                assert np.min(array[:, 2]) == 0
        elif feature == FEATURES_NAMES[1]:
            assert array.shape[0] == len(n_mentions)
        elif feature == FEATURES_NAMES[2]:
            assert array.shape[0] == len(n_mentions)
            assert np.array_equiv(array[:, 0], np.array(list(range(len(n_mentions)))))
        elif feature == FEATURES_NAMES[3]:
            assert array.shape[0] == len(n_mentions)
            assert np.array_equiv(array[:, 0], np.array([p*(p-1)/2 for p in range(len(n_mentions))]))
        elif feature == FEATURES_NAMES[4]:
            assert array.shape[0] == len(n_mentions)
        elif feature == FEATURES_NAMES[5]:
            assert array.shape[0] == len(n_mentions)
        elif feature == FEATURES_NAMES[6]:
            assert array.shape[0] == len(n_mentions)*(len(n_mentions)-1)/2
            assert np.max(array) == len(n_mentions)-2
        elif feature == FEATURES_NAMES[7]:
            if compressed:
                assert array.shape[0] == len(n_mentions)*(len(n_mentions)-1)/2
                assert np.max(array[:, 7]) == len(n_mentions)-2
                assert np.min(array[:, 7]) == 0
        elif feature == FEATURES_NAMES[8]:
            assert array.shape[0] == len(n_mentions)*(len(n_mentions)-1)/2

###############################################################################################
### PARALLEL FCT (has to be at top-level of the module to be pickled for multiprocessing) #####
def load_file(full_name, debug=False):
    '''
    load a *._conll file
    Input: full_name: path to the file
    Output: list of tuples for each conll doc in the file, where the tuple contains:
        (utts_text ([str]): list of the utterances in the document 
         utts_tokens ([[str]]): list of the tokens (conll words) in the document 
         utts_corefs: list of coref objects (dicts) with the following properties:
            coref['label']: id of the coreference cluster,
            coref['start']: start index (index of first token in the utterance),
            coref['end': end index (index of last token in the utterance).
         utts_speakers ([str]): list of the speaker associated to each utterances in the document 
         name (str): name of the document
         part (str): part of the document
        )
    '''
    docs = []
    with io.open(full_name, 'rt', encoding='utf-8', errors='strict') as f:
        lines = list(f)#.readlines()
        utts_text = []
        utts_tokens = []
        utts_corefs = []
        utts_speakers = []
        tokens = []
        corefs = []
        index = 0
        speaker = ""
        name = ""
        part = ""
        for li, line in enumerate(lines):
            cols = line.split()
            if debug: print("line", li, "cols:", cols)
            # End of utterance
            if len(cols) == 0:
                if tokens:
                    if debug: print("End of utterance")
                    utts_text.append(u''.join(t + u' ' for t in tokens))
                    utts_tokens.append(tokens)
                    utts_speakers.append(speaker)
                    utts_corefs.append(corefs)
                    tokens = []
                    corefs = []
                    index = 0
                    speaker = ""
                    continue
            # End of doc
            elif len(cols) == 2:
                if debug: print("End of doc")
                if cols[0] == "#end":
                    if debug: print("Saving doc")
                    docs.append((utts_text, utts_tokens, utts_corefs, utts_speakers, name, part))
                    utts_text = []
                    utts_tokens = []
                    utts_corefs = []
                    utts_speakers = []
                else:
                    raise ValueError("Error on end line " + line)
            # New doc
            elif len(cols) == 5:
                if debug: print("New doc")
                if cols[0] == "#begin":
                    name = re.match(r"\((.*)\);", cols[2]).group(1)
                    try:
                        part = cols[4]
                    except ValueError:
                        print("Error parsing document part " + line)
                    if debug: print("New doc", name, part, name[:2])
                    tokens = []
                    corefs = []
                    index = 0
                else:
                    raise ValueError("Error on begin line " + line)
            # Inside utterance
            elif len(cols) > 7:
                if debug: print("Inside utterance")
                assert (cols[0] == name and int(cols[1]) == int(part)), "Doc name or part error " + line
                assert (int(cols[2]) == index), "Index error on " + line
                if speaker:
                    assert (cols[9] == speaker), "Speaker changed in " + line + speaker
                else:
                    speaker = cols[9]
                    if debug: print("speaker", speaker)
                if cols[-1] != u'-':
                    coref_expr = cols[-1].split(u'|')
                    if debug: print("coref_expr", coref_expr)
                    if not coref_expr:
                        raise ValueError("Coref expression empty " + line)
                    for tok in coref_expr:
                        if debug: print("coref tok", tok)
                        try:
                            match = re.match(r"^(\(?)(\d+)(\)?)$", tok)
                        except:
                            print("error getting coreferences for line " + line)
                        assert match is not None, "Error parsing coref " + tok + " in " + line
                        num = match.group(2)
                        assert (num is not u''), "Error parsing coref " + tok + " in " + line
                        if match.group(1) == u'(':
                            if debug: print("New coref", num)
                            corefs.append({'label': num, 'start': index, 'end': None})
                        if match.group(3) == u')':
                            j = None
                            for i in range(len(corefs)-1, -1, -1):
                                if debug: print("i", i)
                                if corefs[i]['label'] == num and corefs[i]['end'] is None:
                                    j = i
                                    break
                            assert (j is not None), "coref closing error " + line
                            if debug: print("End coref", num)
                            corefs[j]['end'] = index
                tokens.append(clean_token(cols[3]))
                index += 1
            else:
                raise ValueError("Line not standard " + line)
    return docs

def set_feats(doc):
    doc.set_mentions_features()

def get_feats(doc, i):
    return doc.get_feature_array(doc_id=i)

def gather_feats(gathering_array, array, feat_name, pairs_ant_index, pairs_start_index):
    if gathering_array is None:
        gathering_array = array
    else:
        if feat_name == FEATURES_NAMES[6]:
            array = [a + pairs_ant_index for a in array]
        elif feat_name == FEATURES_NAMES[3]:
            array = [a + pairs_start_index for a in array]
        gathering_array += array
    return feat_name, gathering_array

def read_file(full_name):
    doc = ""
    with io.open(full_name, 'rt', encoding='utf-8', errors='strict') as f:
        doc = f.read()
    return doc

###################
### ConllDoc #####

class ConllDoc(Document):
    def __init__(self, name, part, *args, **kwargs):
        self.name = name
        self.part = part
        self.feature_matrix = {}
        self.conll_tokens = []
        self.conll_lookup = []
        self.gold_corefs = []
        self.missed_gold = []
        super(ConllDoc, self).__init__(*args, **kwargs)

    def get_conll_spacy_lookup(self, conll_tokens, spacy_tokens, debug=False):
        '''
        Compute a look up table between spacy tokens (from spacy tokenizer)
        and conll pre-tokenized tokens
        Output: list[conll_index] => list of associated spacy tokens (assume spacy tokenizer has a finer granularity)
        '''
        lookup = []
        c_iter = (t for t in conll_tokens)
        s_iter = enumerate(t for t in spacy_tokens)
        i, s_tok = next(s_iter)
        for c_tok in c_iter:
            #if debug: print("conll", c_tok, "spacy", s_tok, "index", i)
            c_lookup = []
            while i is not None and len(c_tok) and c_tok.startswith(s_tok.text):
                c_lookup.append(i)
                c_tok = c_tok[len(s_tok):]
                i, s_tok = next(s_iter, (None, None))
                if debug and len(c_tok): print("eating token: conll", c_tok, "spacy", s_tok, "index", i)
            assert len(c_lookup), "Unmatched conll and spacy tokens"
            lookup.append(c_lookup)
        return lookup

    def add_conll_utterance(self, parsed, tokens, corefs, speaker_id, use_gold_mentions=False, debug=False):
        conll_lookup = self.get_conll_spacy_lookup(tokens, parsed)
        self.conll_tokens.append(tokens)
        self.conll_lookup.append(conll_lookup)
        # Convert conll tokens coref index in spacy tokens indexes
        identified_gold = [False] * len(corefs)
        for coref in corefs:
            assert (coref['label'] is not None and coref['start'] is not None and coref['end'] is not None), \
                ("Error in coreference " + coref + " in " + parsed)
            coref['start'] = conll_lookup[coref['start']][0]
            coref['end'] = conll_lookup[coref['end']][-1]

        if speaker_id not in self.speakers:
            speaker_name = speaker_id.split(u'_')
            if debug: print("New speaker: ", speaker_id, "name: ", speaker_name)
            self.speakers[speaker_id] = Speaker(speaker_id, speaker_name)
        if use_gold_mentions:
            for coref in corefs:
                #            print("coref['label']", coref['label'])
                #            print("coref text",parsed[coref['start']:coref['end']])
                mention = Mention(parsed[coref['start']:coref['end']], len(self.mentions), len(self.utterances),
                                  self.n_sents, speaker=self.speakers[speaker_id], gold_label=coref['label'])
                self.mentions.append(mention)
                #            print("mention: ", mention, "label", mention.gold_label)
        else:
            self._extract_mentions(parsed, len(self.utterances), self.n_sents, self.speakers[speaker_id])
            # Assign a gold label to mentions which have one
            if debug: print("Check corefs", corefs)
            for i, coref in enumerate(corefs):
                for m in self.mentions:
                    if m.utterance_index != len(self.utterances):
                        continue
                    # if debug: print("Checking mention", m, m.utterance_index, m.start, m.end)
                    if coref['start'] == m.start and coref['end'] == m.end - 1:
                        m.gold_label = coref['label']
                        identified_gold[i] = True
                        # if debug: print("Gold mention found:", m, coref['label'])
            for found, coref in zip(identified_gold, corefs):
                if not found:
                    self.missed_gold.append([self.name, self.part, str(len(self.utterances)), parsed.text, parsed[coref['start']:coref['end']+1].text])
                    if debug:
                        print("❄️ gold mention not in predicted mentions", coref, parsed[coref['start']:coref['end']+1])
        self.utterances.append(parsed)
        self.gold_corefs.append(corefs)
        self.utterances_speaker.append(self.speakers[speaker_id])
        self.n_sents += len(list(parsed.sents))

    def get_single_mention_features_conll(self, mention, compressed=True):
        ''' Compressed or not single mention features'''
        if not compressed:
            _, features = self.get_single_mention_features(mention)
            return features[np.newaxis, :]
        feat_l = [mention.features_["01_MentionType"],
                  mention.features_["02_MentionLength"],
                  mention.index,
                  len(self.mentions),
                  mention.features_["04_IsMentionNested"],
                  self.genre_,
                 ]
        return feat_l

    def get_pair_mentions_features_conll(self, m1, m2, compressed=True):
        ''' Compressed or not single mention features'''
        if not compressed:
            _, features = self.get_pair_mentions_features(m1, m2)
            return features[np.newaxis, :]
        features_, _ = self.get_pair_mentions_features(m1, m2)
        feat_l = [features_["00_SameSpeaker"],
                  features_["01_AntMatchMentionSpeaker"],
                  features_["02_MentionMatchSpeaker"],
                  features_["03_HeadsAgree"],
                  features_["04_ExactStringMatch"],
                  features_["05_RelaxedStringMatch"],
                  features_["06_SentenceDistance"],
                  features_["07_MentionDistance"],
                  features_["08_Overlapping"],
                 ]
        return feat_l

    def get_feature_array(self, doc_id, feature=None, compressed=True, debug=True):
        """
        Prepare feature array:
            mentions_spans: (N, S)
            mentions_words: (N, W)
            mentions_features: (N, Fs)
            mentions_labels: (N, 1)
            mentions_pairs_start_index: (N, 1) index of beggining of pair list in pair_labels
            mentions_pairs_length: (N, 1) number of pairs (i.e. nb of antecedents) for each mention 
            pairs_features: (P, Fp)
            pairs_labels: (P, 1)
            pairs_ant_idx: (P, 1) => indexes of antecedents mention for each pair (mention index in doc)
        """ 
        if not self.mentions:
            print("No mention in this doc !")
            return {}
        if debug: print("🛎 features matrices")
        mentions_spans = []
        mentions_words = []
        mentions_features = []
        pairs_ant_idx = []
        pairs_features = []
        pairs_labels = []
        mentions_labels = []
        mentions_pairs_start = []
        mentions_pairs_length = []
        mentions_location = []
        n_mentions = 0
        total_pairs = 0
        if debug: print("mentions", self.mentions, str([m.gold_label for m in self.mentions]))
        for mention_idx, antecedents_idx in list(self.get_candidate_pairs(max_distance=None, max_distance_with_match=None)):
            n_mentions += 1
            mention = self.mentions[mention_idx]
            mentions_spans.append(mention.spans_embeddings)
            w_idx = mention_words_idx(self.embed_extractor, mention)
            if w_idx is None:
                print("error in", self.name, self.part, mention.utterance_index)
            mentions_words.append(w_idx)
            mentions_features.append(self.get_single_mention_features_conll(mention, compressed))
            mentions_location.append([mention.start, mention.end, mention.utterance_index, mention_idx, doc_id])
            ants = [self.mentions[ant_idx] for ant_idx in antecedents_idx]
            no_antecedent = not any(ant.gold_label == mention.gold_label for ant in ants) or mention.gold_label is None
            if antecedents_idx:
                pairs_ant_idx += [idx for idx in antecedents_idx]
                pairs_features += [self.get_pair_mentions_features_conll(ant, mention, compressed) for ant in ants]
                ant_labels = [0 for ant in ants] if no_antecedent else [1 if ant.gold_label == mention.gold_label else 0 for ant in ants]
                pairs_labels += ant_labels
            mentions_labels.append(1 if no_antecedent else 0)
            mentions_pairs_start.append(total_pairs)
            total_pairs += len(ants)
            mentions_pairs_length.append(len(ants))

        out_dict = {FEATURES_NAMES[0]: mentions_features,
                    FEATURES_NAMES[1]: mentions_labels,
                    FEATURES_NAMES[2]: mentions_pairs_length,
                    FEATURES_NAMES[3]: mentions_pairs_start,
                    FEATURES_NAMES[4]: mentions_spans,
                    FEATURES_NAMES[5]: mentions_words,
                    FEATURES_NAMES[6]: pairs_ant_idx if pairs_ant_idx else None,
                    FEATURES_NAMES[7]: pairs_features if pairs_features else None,
                    FEATURES_NAMES[8]: pairs_labels if pairs_labels else None,
                    FEATURES_NAMES[9]: [mentions_location],
                    FEATURES_NAMES[10]: [self.conll_tokens],
                    FEATURES_NAMES[11]: [self.conll_lookup],
                    FEATURES_NAMES[12]: [{'name': self.name,
                                          'part': self.part,
                                          'utterances': list(str(u) for u in self.utterances),
                                          'mentions': list(str(m) for m in self.mentions)}],
                    }
        if debug:
            print("🚘 Summary")
            for k, v in out_dict.items():
                print(k, len(v))
        return n_mentions, total_pairs, out_dict

###################
### ConllCorpus #####
class ConllCorpus(object):
    def __init__(self, n_jobs=4, embed_path=PACKAGE_DIRECTORY+"/weights/", use_gold_mentions=False):
        self.n_jobs = n_jobs
        self.features = {}
        self.utts_text = []
        self.utts_tokens = []
        self.utts_corefs = []
        self.utts_speakers = []
        self.utts_doc_idx = []
        self.docs_names = []
        self.docs = []
        if embed_path is not None:
            self.embed_extractor = EmbeddingExtractor(embed_path)
        self.trainable_embed = []
        self.trainable_voc = []
        self.use_gold_mentions = use_gold_mentions

    def check_words_in_embeddings_voc(self, embedding, tuned=True, debug=False):
        print("🌋 Checking if words are in embedding voc")
        if tuned:
            embed_voc = embedding.tun_idx
        else:
            embed_voc = embedding.stat_idx
        missing_words = []
        missing_words_sents = []
        missing_words_doc = []
        for doc in self.docs:
            # if debug: print("Checking doc", doc.name, doc.part)
            for sent in doc.utterances:
                # if debug: print(sent.text)
                for word in sent:
                    w = embedding.normalize_word(word)
                    # if debug: print(w)
                    if w not in embed_voc:
                        missing_words.append(w)
                        missing_words_sents.append(sent.text)
                        missing_words_doc.append(doc.name + doc.part)
                        if debug:
                            out_str = "No matching tokens in tuned voc for " + w + \
                                      " in sentence " + sent.text + \
                                      " in doc " + doc.name + doc.part
                            print(out_str)
        return missing_words, missing_words_sents, missing_words_doc

    def test_sentences_words(self, save_file, debug=False):
        print("🌋 Saving sentence list")
        with io.open(save_file, "w", encoding='utf-8') as f:
            if debug: print("Sentences saved in", save_file)
            for doc in self.docs:
                out_str = "#begin document (" + doc.name + \
                          "); part " + doc.part + "\n"
                f.write(out_str)
                for sent in doc.utterances:
                    f.write(sent.text + '\n')
                out_str = "#end document\n\n"
                f.write(out_str)

    def save_sentences(self, save_file, debug=False):
        print("🌋 Saving sentence list")
        with io.open(save_file, "w", encoding='utf-8') as f:
            if debug: print("Sentences saved in", save_file)
            for doc in self.docs:
                out_str = "#begin document (" + doc.name + \
                          "); part " + doc.part + "\n"
                f.write(out_str)
                for sent in doc.utterances:
                    f.write(sent.text + '\n')
                out_str = "#end document\n\n"
                f.write(out_str)

    def build_key_file(self, data_path, key_file, debug=False):
        print("🌋 Building key file from corpus")
        print("Saving in", key_file)
        # Create a pool of processes. By default, one is created for each CPU in your machine.
        with io.open(key_file, "w", encoding='utf-8') as kf:
            if debug: print("Key file saved in", key_file)
            for dirpath, _, filenames in os.walk(data_path):
                print("In", dirpath)
                file_list = [os.path.join(dirpath, f) for f in filenames if f.endswith(".v4_auto_conll") \
                            or f.endswith(".v4_gold_conll")]
                cleaned_file_list = []
                for f in file_list:
                    fn = f.split('.')
                    if fn[1] == "v4_auto_conll":
                        gold = fn[0] + "." + "v4_gold_conll"
                        if gold not in file_list:
                            cleaned_file_list.append(f)
                    else:
                        cleaned_file_list.append(f)
            #self.load_file(file_list[0])
                doc_list = parallel_process(cleaned_file_list, read_file)
                for doc in doc_list:
                    kf.write(doc)

    def list_undetected_mentions(self, data_path, save_file, debug=True):
        self.read_corpus(data_path)
        print("🌋 Listing undetected mentions")
        with io.open(save_file, 'w', encoding='utf-8') as out_file:
            for doc in tqdm(self.docs):
                for name, part, utt_i, utt, coref in doc.missed_gold:
                    out_str = name + u"\t" + part + u"\t" + utt_i + u'\t"' + utt + u'"\n'
                    out_str += coref + u"\n"
                    out_file.write(out_str)
                    if debug: print(out_str)

    def read_corpus(self, data_path, debug=False):
        print("🌋 Reading files")
        for dirpath, _, filenames in os.walk(data_path):
            print("In", dirpath, os.path.abspath(dirpath))
            file_list = [os.path.join(dirpath, f) for f in filenames if f.endswith(".v4_auto_conll") \
                        or f.endswith(".v4_gold_conll")]
            cleaned_file_list = []
            for f in file_list:
                fn = f.split('.')
                if fn[1] == "v4_auto_conll":
                    gold = fn[0] + "." + "v4_gold_conll"
                    if gold not in file_list:
                        cleaned_file_list.append(f)
                else:
                    cleaned_file_list.append(f)
            doc_list = parallel_process(cleaned_file_list, load_file)
            for docs in doc_list:#executor.map(self.load_file, cleaned_file_list):
                for utts_text, utt_tokens, utts_corefs, utts_speakers, name, part in docs:
                    print("Imported", name)
                    if debug:
                        print("utts_text", utts_text)
                        print("utt_tokens", utt_tokens)
                        print("utts_corefs", utts_corefs)
                        print("utts_speakers", utts_speakers)
                        print("name, part", name, part)
                    self.utts_text += utts_text
                    self.utts_tokens += utt_tokens
                    self.utts_corefs += utts_corefs
                    self.utts_speakers += utts_speakers
                    self.utts_doc_idx += [len(self.docs_names)] * len(utts_text)
                    self.docs_names.append((name, part))
        print("utts_text size", len(self.utts_text))
        print("utts_tokens size", len(self.utts_tokens))
        print("utts_corefs size", len(self.utts_corefs))
        print("utts_speakers size", len(self.utts_speakers))
        print("utts_doc_idx size", len(self.utts_doc_idx))
        print("🌋 Building docs")
        for name, part in self.docs_names:
            self.docs.append(ConllDoc(name=name, part=part, nlp=None,
                                      blacklist=False, consider_speakers=True,
                                      embedding_extractor=self.embed_extractor,
                                      conll=CONLL_GENRES[name[:2]]))
        print("🌋 Loading spacy model")
        try:
            spacy.info('en_core_web_sm')
            model = 'en_core_web_sm'
        except IOError:
            print("No spacy 2 model detected, using spacy1 'en' model")
            spacy.info('en')
            model = 'en'
        nlp = spacy.load(model)
        print("🌋 Parsing utterances and filling docs")
        doc_iter = (s for s in self.utts_text)
        for utt_tuple in tqdm(zip(nlp.pipe(doc_iter),
                                           self.utts_tokens, self.utts_corefs,
                                           self.utts_speakers, self.utts_doc_idx)):
            spacy_tokens, conll_tokens, corefs, speaker, doc_id = utt_tuple
            if debug: print(unicode_(self.docs_names[doc_id]), "-", spacy_tokens)
            doc = spacy_tokens
            if debug: 
                out_str = "utterance " + unicode_(doc) + " corefs " + unicode_(corefs) + \
                          " speaker " + unicode_(speaker) + "doc_id" + unicode_(doc_id)
                print(out_str.encode('utf-8'))
            self.docs[doc_id].add_conll_utterance(doc, conll_tokens, corefs, speaker,
                                                  use_gold_mentions=self.use_gold_mentions)

    def build_and_gather_multiple_arrays(self, save_path):
        print("🌋 Extracting mentions features")
        parallel_process(self.docs, set_feats, n_jobs=self.n_jobs)

        print("🌋 Building and gathering arrays")
        arr =[{'doc': doc,
               'i': i} for i, doc in enumerate(self.docs)]
        arrays_dicts = parallel_process(arr, get_feats, use_kwargs=True, n_jobs=self.n_jobs)
        gathering_dict = dict((feat, None) for feat in FEATURES_NAMES)
        n_mentions_list = []
        pairs_ant_index = 0
        pairs_start_index = 0
        for n, p, arrays_dict in tqdm(arrays_dicts):
            for f in FEATURES_NAMES:
                if gathering_dict[f] is None:
                    gathering_dict[f] = arrays_dict[f]
                else:
                    if f == FEATURES_NAMES[6]:
                        array = [a + pairs_ant_index for a in arrays_dict[f]]
                    elif f == FEATURES_NAMES[3]:
                        array = [a + pairs_start_index for a in arrays_dict[f]]
                    else:
                        array = arrays_dict[f]
                    gathering_dict[f] += array
            pairs_ant_index += n
            pairs_start_index += p
            n_mentions_list.append(n)

        for feature in FEATURES_NAMES[:9]:
            print("Building numpy array for", feature, "length", len(gathering_dict[feature]))
            if feature != "mentions_spans":
                array = np.array(gathering_dict[feature])
                if array.ndim == 1:
                    array = np.expand_dims(array, axis=1)
            else:
                array = np.stack(gathering_dict[feature])
            # check_numpy_array(feature, array, n_mentions_list)
            print("Saving numpy", feature, "size", array.shape)
            np.save(save_path + feature, array)
        for feature in FEATURES_NAMES[9:]:
            print("Saving pickle", feature, "size", len(gathering_dict[feature]))
            with open(save_path + feature + '.bin', "wb") as fp:  
                pickle.dump(gathering_dict[feature], fp)

    def save_vocabulary(self, save_path, debug=False):
        def _vocabulary_to_file(path, vocabulary):
            print("🌋 Saving vocabulary")
            with io.open(path, "w", encoding='utf-8') as f:
                if debug:
                    print('voc saved in {path}, length: {length}'
                          .format(path=path, size=len(vocabulary)))
                for w in tunable_voc:
                    f.write(w + '\n')

        print("🌋 Building tunable vocabulary matrix from static vocabulary")
        tunable_voc = self.embed_extractor.tun_voc
        _vocabulary_to_file(
            path=save_path + 'tuned_word_vocabulary.txt',
            vocabulary=tunable_voc
        )

        static_voc = self.embed_extractor.stat_voc
        _vocabulary_to_file(
            path=save_path + 'static_word_vocabulary.txt',
            vocabulary=static_voc
        )

        tuned_word_embeddings = np.vstack([self.embed_extractor.get_stat_word(w)[1] for w in tunable_voc])
        print("Saving tunable voc, size:", tuned_word_embeddings.shape)
        np.save(save_path + "tuned_word_embeddings", tuned_word_embeddings)

        static_word_embeddings = np.vstack([self.embed_extractor.static_embeddings[w] for w in static_voc])
        print("Saving static voc, size:", static_word_embeddings.shape)
        np.save(save_path + "static_word_embeddings", static_word_embeddings)


if __name__ == '__main__':
    DIR_PATH = os.path.dirname(os.path.realpath(__file__))
    parser = argparse.ArgumentParser(description='Training the neural coreference model')
    parser.add_argument('--function', type=str, default='all', help='Function ("all", "key", "parse", "find_undetected")')
    parser.add_argument('--path', type=str, default=DIR_PATH + '/data/', help='Path to the dataset')
    parser.add_argument('--key', type=str, help='Path to an optional key file for scoring')
    parser.add_argument('--n_jobs', type=int, default=1, help='Number of parallel jobs (default 1)')
    args = parser.parse_args()
    if args.key is None:
        args.key = args.path + "/key.txt"
    CORPUS = ConllCorpus(n_jobs=args.n_jobs)
    if args.function == 'parse' or args.function == 'all':
        SAVE_DIR = args.path + "/numpy/"
        if not os.path.exists(SAVE_DIR):
            os.makedirs(SAVE_DIR)
        else:
            if os.listdir(SAVE_DIR):
                print("There are already data in", SAVE_DIR)
                print("Erasing")
                for file in os.listdir(SAVE_DIR):
                    print(file)
                    os.remove(SAVE_DIR + file)
        start_time = time.time()
        CORPUS.read_corpus(args.path)
        print('=> read_corpus time elapsed', time.time() - start_time)
        start_time2 = time.time()
        CORPUS.build_and_gather_multiple_arrays(SAVE_DIR)
        print('=> build_and_gather_multiple_arrays time elapsed', time.time() - start_time2)
        start_time2 = time.time()
        CORPUS.save_vocabulary(SAVE_DIR)
        print('=> save_vocabulary time elapsed', time.time() - start_time2)
        print('=> total time elapsed', time.time() - start_time)
    if args.function == 'key' or args.function == 'all':
        CORPUS.build_key_file(args.path, args.key)
    if args.function == 'find_undetected':
        CORPUS.list_undetected_mentions(args.path, args.path + "/undetected_mentions.txt")



# coding: utf8
"""Utils"""

from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

from concurrent.futures import ThreadPoolExecutor, as_completed
import os
import numpy as np
from tqdm import tqdm

PACKAGE_DIRECTORY = os.path.dirname(os.path.abspath(__file__))
BATCH_SIZE_PATH = os.path.join(PACKAGE_DIRECTORY, "test_batch_size.txt") #fernandes.txt")#

SIZE_SPAN = 250 # size of the span vector (averaged word embeddings)
SIZE_WORD = 8 # number of words in a mention (tuned embeddings)
SIZE_EMBEDDING = 50 # size of the words embeddings
SIZE_FP = 70 # number of features for a pair of mention
SIZE_FP_COMPRESSED = 9 # size of the features for a pair of mentions as stored in numpy arrays
SIZE_FS = 24 # number of features of a single mention
SIZE_FS_COMPRESSED = 6 # size of the features for a mention as stored in numpy arrays
SIZE_GENRE = 7 # Size of the genre one-hot array
SIZE_MENTION_EMBEDDING = SIZE_SPAN + SIZE_WORD * SIZE_EMBEDDING # A mention embeddings (span + words vectors)
SIZE_SNGL_FEATS = SIZE_FS - SIZE_GENRE
SIZE_PAIR_FEATS = SIZE_FP - SIZE_GENRE
SIZE_SNGL_IN_NO_GENRE = SIZE_MENTION_EMBEDDING + SIZE_SNGL_FEATS
SIZE_PAIR_IN_NO_GENRE = 2 * SIZE_MENTION_EMBEDDING + SIZE_PAIR_FEATS

SIZE_PAIR_IN = 2 * SIZE_MENTION_EMBEDDING + SIZE_FP # Input to the mentions pair neural network
SIZE_SINGLE_IN = SIZE_MENTION_EMBEDDING + SIZE_FS  # Input to the single mention neural network

DISTANCE_BINS = list(range(5)) + [5]*3 + [6]*8 + [7]*16 +[8]*32
BINS_NUM = float(len(DISTANCE_BINS))
MAX_BINS = DISTANCE_BINS[-1] + 1

def encode_distance(x):
    ''' Encode an integer or an array of integers as a (bined) one-hot numpy array '''
    def _encode_distance(d):
        ''' Encode an integer as a (bined) one-hot numpy array '''
        dist_vect = np.zeros((11,), dtype='float32')
        if d < 64:
            dist_vect[DISTANCE_BINS[d]] = 1
        else:
            dist_vect[9] = 1
        dist_vect[10] = min(float(d), BINS_NUM) / BINS_NUM
        return dist_vect

    if isinstance(x, np.ndarray):
        arr_l = [_encode_distance(y)[np.newaxis, :] for y in x]
        out_arr = np.concatenate(arr_l)
    else:
        out_arr = _encode_distance(x)
    return out_arr

def parallel_process(array, function, n_jobs=16, use_kwargs=False, front_num=10):
    """
        A parallel version of the map function with a progress bar. 

        Args:
            array (array-like): An array to iterate over.
            function (function): A python function to apply to the elements of array
            n_jobs (int, default=16): The number of cores to use
            use_kwargs (boolean, default=False): Whether to consider the elements of array as dictionaries of 
                keyword arguments to function 
            front_num (int, default=3): The number of iterations to run serially before kicking off the parallel job. 
                Useful for catching bugs
        Returns:
            [function(array[0]), function(array[1]), ...]
    """
    #We run the first few iterations serially to catch bugs
    if front_num > 0:
        front = [function(**a) if use_kwargs else function(a) for a in array[:front_num]]
    else:
        front = []
    #If we set n_jobs to 1, just run a list comprehension. This is useful for benchmarking and debugging.
    if n_jobs==1:
        return front + [function(**a) if use_kwargs else function(a) for a in tqdm(array[front_num:])]
    #Assemble the workers
    with ThreadPoolExecutor(max_workers=n_jobs) as pool:
        #Pass the elements of array into function
        if use_kwargs:
            futures = [pool.submit(function, **a) for a in array[front_num:]]
        else:
            futures = [pool.submit(function, a) for a in array[front_num:]]
        kwargs = {
            'total': len(futures),
            'unit': 'it',
            'unit_scale': True,
            'leave': True
        }
        # #Print out the progress as tasks complete
        # for _ in tqdm(as_completed(futures), **kwargs):
        #     pass
    out = []
    #Get the results from the futures. 
    for future in futures: #tqdm(futures):
        try:
            out.append(future.result())
        except Exception as e:
            out.append(e)
    return front + out



# coding: utf8
# cython: profile=True
# cython: infer_types=True
"""Coref resolution"""

from __future__ import unicode_literals
from __future__ import print_function

import sys
import os
import spacy
import numpy as np

from neuralcoref.utils import PACKAGE_DIRECTORY, SIZE_PAIR_IN, SIZE_SINGLE_IN
from neuralcoref.compat import unicode_
from neuralcoref.document import Document, MENTION_TYPE, NO_COREF_LIST

#######################
##### UTILITIES #######

MAX_FOLLOW_UP = 50

#######################
###### CLASSES ########

class Model(object):
    '''
    Coreference neural model
    '''
    def __init__(self, model_path):
        weights, biases = [], []
        for file in sorted(os.listdir(model_path)):
            if file.startswith("single_mention_weights"):
                w = np.load(os.path.join(model_path, file))
                weights.append(w)
            if file.startswith("single_mention_bias"):
                w = np.load(os.path.join(model_path, file))
                biases.append(w)
        self.single_mention_model = list(zip(weights, biases))
        weights, biases = [], []
        for file in sorted(os.listdir(model_path)):
            if file.startswith("pair_mentions_weights"):
                w = np.load(os.path.join(model_path, file))
                weights.append(w)
            if file.startswith("pair_mentions_bias"):
                w = np.load(os.path.join(model_path, file))
                biases.append(w)
        self.pair_mentions_model = list(zip(weights, biases))

    def _score(self, features, layers):
        for weights, bias in layers:
            #print("features", features.shape)
            features = np.matmul(weights, features) + bias
            if weights.shape[0] > 1:
                features = np.maximum(features, 0) # ReLU
        return np.sum(features, axis=0)

    def get_multiple_single_score(self, first_layer_input):
        return self._score(first_layer_input, self.single_mention_model)

    def get_multiple_pair_score(self, first_layer_input):
        return self._score(first_layer_input, self.pair_mentions_model)


class Coref(object):
    '''
    Main coreference resolution algorithm
    '''
    def __init__(self, nlp=None, greedyness=0.5, max_dist=50, max_dist_match=500, conll=None,
                 blacklist=True, debug=False):
        self.greedyness = greedyness
        self.max_dist = max_dist
        self.max_dist_match = max_dist_match
        self.debug = debug
        model_path = os.path.join(PACKAGE_DIRECTORY, "weights/conll/" if conll is not None else "weights/")
        model_path = os.path.join(PACKAGE_DIRECTORY, "weights/")
        print("Loading neuralcoref model from", model_path)
        self.coref_model = Model(model_path)
        if nlp is None:
            print("Loading spacy model")
            try:
                spacy.info('en_core_web_sm')
                model = 'en_core_web_sm'
            except IOError:
                print("No spacy 2 model detected, using spacy1 'en' model")
                spacy.info('en')
                model = 'en'
            nlp = spacy.load(model)
        self.data = Document(nlp, conll=conll, blacklist=blacklist, model_path=model_path)
        self.clusters = {}
        self.mention_to_cluster = []
        self.mentions_single_scores = {}
        self.mentions_pairs_scores = {}

    ###################################
    #### ENTITY CLUSTERS FUNCTIONS ####
    ###################################

    def _prepare_clusters(self):
        '''
        Clean up and prepare one cluster for each mention
        '''
        self.mention_to_cluster = list(range(len(self.data.mentions)))
        self.clusters = dict((i, [i]) for i in self.mention_to_cluster)
        self.mentions_single_scores = {}
        self.mentions_pairs_scores = {}
        for mention in self.mention_to_cluster:
            self.mentions_single_scores[mention] = None
            self.mentions_pairs_scores[mention] = {}

    def _merge_coreference_clusters(self, ant_idx, mention_idx):
        '''
        Merge two clusters together
        '''
        if self.mention_to_cluster[ant_idx] == self.mention_to_cluster[mention_idx]:
            return

        remove_id = self.mention_to_cluster[ant_idx]
        keep_id = self.mention_to_cluster[mention_idx]
        for idx in self.clusters[remove_id]:
            self.mention_to_cluster[idx] = keep_id
            self.clusters[keep_id].append(idx)

        del self.clusters[remove_id]

    def remove_singletons_clusters(self):
        remove_id = []
        for key, mentions in self.clusters.items():
            if len(mentions) == 1:
                remove_id.append(key)
                self.mention_to_cluster[key] = None
        for rem in remove_id:
            del self.clusters[rem]

    def display_clusters(self):
        '''
        Print clusters informations
        '''
        print(self.clusters)
        for key, mentions in self.clusters.items():
            print("cluster", key, "(", ", ".join(unicode_(self.data[m]) for m in mentions), ")")

    ###################################
    ####### MAIN COREF FUNCTIONS ######
    ###################################

    def run_coref_on_mentions(self, mentions):
        '''
        Run the coreference model on a mentions list
        '''
        best_ant = {}
        best_score = {}
        n_ant = 0
        inp = np.empty((SIZE_SINGLE_IN, len(mentions)))
        for i, mention_idx in enumerate(mentions):
            mention = self.data[mention_idx]
            print("mention.embedding", mention.embedding.shape)
            inp[:len(mention.embedding), i] = mention.embedding
            inp[:len(mention.embedding), i] = mention.features
            inp[:len(mention.embedding), i] = self.data.genre
        score = self.coref_model.get_multiple_single_score(inp.T)
        for mention_idx, s in zip(mentions, score):
            self.mentions_single_scores[mention_idx] = s
            best_score[mention_idx] = s - 50 * (self.greedyness - 0.5)

        for mention_idx, ant_list in self.data.get_candidate_pairs(mentions, self.max_dist, self.max_dist_match):
            if len(ant_list) == 0:
                continue
            inp_l = []
            for ant_idx in ant_list:
                mention = self.data[mention_idx]
                antecedent = self.data[ant_idx]
                feats_, pwf = self.data.get_pair_mentions_features(antecedent, mention)
                inp_l.append(pwf)
            inp = np.stack(inp_l, axis=0)
            score = self.coref_model.get_multiple_pair_score(inp.T)
            for ant_idx, s in zip(ant_list, score):
                self.mentions_pairs_scores[mention_idx][ant_idx] = s
                if s > best_score[mention_idx]:
                    best_score[mention_idx] = s
                    best_ant[mention_idx] = ant_idx
            if mention_idx in best_ant:
                n_ant += 1
                self._merge_coreference_clusters(best_ant[mention_idx], mention_idx)
        return (n_ant, best_ant)

    def run_coref_on_utterances(self, last_utterances_added=False, follow_chains=True, debug=False):
        ''' Run the coreference model on some utterances

        Arg:
            last_utterances_added: run the coreference model over the last utterances added to the data
            follow_chains: follow coreference chains over previous utterances
        '''
        if debug: print("== run_coref_on_utterances == start")
        self._prepare_clusters()
        if debug: self.display_clusters()
        mentions = list(self.data.get_candidate_mentions(last_utterances_added=last_utterances_added))
        n_ant, antecedents = self.run_coref_on_mentions(mentions)
        mentions = antecedents.values()
        if follow_chains and last_utterances_added and n_ant > 0:
            i = 0
            while i < MAX_FOLLOW_UP:
                i += 1
                n_ant, antecedents = self.run_coref_on_mentions(mentions)
                mentions = antecedents.values()
                if n_ant == 0:
                    break
        if debug: self.display_clusters()
        if debug: print("== run_coref_on_utterances == end")

    def one_shot_coref(self, utterances, utterances_speakers_id=None, context=None,
                       context_speakers_id=None, speakers_names=None):
        ''' Clear history, load a list of utterances and an optional context and run the coreference model on them

        Arg:
        - `utterances` : iterator or list of string corresponding to successive utterances (in a dialogue) or sentences.
            Can be a single string for non-dialogue text.
        - `utterances_speakers_id=None` : iterator or list of speaker id for each utterance (in the case of a dialogue).
            - if not provided, assume two speakers speaking alternatively.
            - if utterances and utterances_speaker are not of the same length padded with None
        - `context=None` : iterator or list of string corresponding to additionnal utterances/sentences sent prior to `utterances`. Coreferences are not computed for the mentions identified in `context`. The mentions in `context` are only used as possible antecedents to mentions in `uterrance`. Reduce the computations when we are only interested in resolving coreference in the last sentences/utterances.
        - `context_speakers_id=None` : same as `utterances_speakers_id` for `context`. 
        - `speakers_names=None` : dictionnary of list of acceptable speaker names (strings) for speaker_id in `utterances_speakers_id` and `context_speakers_id`
        Return:
            clusters of entities with coreference resolved
        '''
        self.data.set_utterances(context, context_speakers_id, speakers_names)
        self.continuous_coref(utterances, utterances_speakers_id, speakers_names)
        return self.get_clusters()

    def continuous_coref(self, utterances, utterances_speakers_id=None, speakers_names=None):
        '''
        Only resolve coreferences for the mentions in the utterances
        (but use the mentions in previously loaded utterances as possible antecedents)
        Arg:
            utterances : iterator or list of string corresponding to successive utterances
            utterances_speaker : iterator or list of speaker id for each utterance.
                If not provided, assume two speakers speaking alternatively.
                if utterances and utterances_speaker are not of the same length padded with None
            speakers_names : dictionnary of list of acceptable speaker names for each speaker id
        Return:
            clusters of entities with coreference resolved
        '''
        self.data.add_utterances(utterances, utterances_speakers_id, speakers_names)
        self.run_coref_on_utterances(last_utterances_added=True, follow_chains=True)
        return self.get_clusters()

    ###################################
    ###### INFORMATION RETRIEVAL ######
    ###################################

    def get_utterances(self, last_utterances_added=True):
        ''' Retrieve the list of parsed uterrances'''
        if last_utterances_added and len(self.data.last_utterances_loaded):
            return [self.data.utterances[idx] for idx in self.data.last_utterances_loaded]
        else:
            return self.data.utterances

    def get_resolved_utterances(self, last_utterances_added=True, blacklist=True):
        ''' Return a list of utterrances text where the '''
        coreferences = self.get_most_representative(last_utterances_added, blacklist)
        resolved_utterances = []
        for utt in self.get_utterances(last_utterances_added=last_utterances_added):
            resolved_utt = ""
            in_coref = None
            for token in utt:
                if in_coref is None:
                    for coref_original, coref_replace in coreferences.items():
                        if coref_original[0] == token:
                            in_coref = coref_original
                            resolved_utt += coref_replace.text.lower()
                            break
                    if in_coref is None:
                        resolved_utt += token.text_with_ws
                if in_coref is not None and token == in_coref[-1]:
                    resolved_utt += ' ' if token.whitespace_ and resolved_utt[-1] is not ' ' else ''
                    in_coref = None
            resolved_utterances.append(resolved_utt)
        return resolved_utterances

    def get_mentions(self):
        ''' Retrieve the list of mentions'''
        return self.data.mentions

    def get_scores(self):
        ''' Retrieve scores for single mentions and pair of mentions'''
        return {"single_scores": self.mentions_single_scores,
                "pair_scores": self.mentions_pairs_scores}

    def get_clusters(self, remove_singletons=False, blacklist=False):
        ''' Retrieve cleaned clusters'''
        clusters = self.clusters
        mention_to_cluster = self.mention_to_cluster
        remove_id = []
        if blacklist:
            for key, mentions in clusters.items():
                cleaned_list = []
                for mention_idx in mentions:
                    mention = self.data.mentions[mention_idx]
                    if mention.lower_ not in NO_COREF_LIST:
                        cleaned_list.append(mention_idx)
                clusters[key] = cleaned_list
            # Also clean up keys so we can build coref chains in self.get_most_representative
            added = {}
            for key, mentions in clusters.items():
                if self.data.mentions[key].lower_ in NO_COREF_LIST:
                    remove_id.append(key)
                    mention_to_cluster[key] = None
                    if mentions:
                        added[mentions[0]] = mentions
            for rem in remove_id:
                del clusters[rem]
            clusters.update(added)

        if remove_singletons:
            remove_id = []
            for key, mentions in clusters.items():
                if len(mentions) == 1:
                    remove_id.append(key)
                    mention_to_cluster[key] = None
            for rem in remove_id:
                del clusters[rem]

        return clusters, mention_to_cluster

    def get_most_representative(self, last_utterances_added=True, blacklist=True):
        '''
        Find a most representative mention for each cluster

        Return:
            Dictionnary of {original_mention: most_representative_resolved_mention, ...}
        '''
        clusters, _ = self.get_clusters(remove_singletons=True, blacklist=blacklist)
        coreferences = {}
        for key in self.data.get_candidate_mentions(last_utterances_added=last_utterances_added):
            if self.mention_to_cluster[key] is None:
                continue
            mentions = clusters.get(self.mention_to_cluster[key], None)
            if mentions is None:
                continue
            representative = self.data.mentions[key]
            for mention_idx in mentions[1:]:
                mention = self.data.mentions[mention_idx]
                if mention.mention_type is not representative.mention_type:
                    if mention.mention_type == MENTION_TYPE["PROPER"] \
                        or (mention.mention_type == MENTION_TYPE["NOMINAL"] and
                                representative.mention_type == MENTION_TYPE["PRONOMINAL"]):
                        coreferences[self.data.mentions[key]] = mention
                        representative = mention

        return coreferences



# coding: utf8
from __future__ import unicode_literals, absolute_import

import os
import shutil
import tarfile
import tempfile
import logging

# Filter Cython warnings that would force everybody to re-compile from source (like https://github.com/numpy/numpy/pull/432).
import warnings
warnings.filterwarnings("ignore", message="spacy.strings.StringStore size changed,")

from .neuralcoref import NeuralCoref
from .file_utils import NEURALCOREF_MODEL_URL, NEURALCOREF_MODEL_PATH, NEURALCOREF_CACHE, cached_path

__all__ = ['NeuralCoref', 'add_to_pipe']
__version__ = "4.0.0"

logger = logging.getLogger(__name__)

if os.path.exists(NEURALCOREF_MODEL_PATH) and os.path.exists(os.path.join(NEURALCOREF_MODEL_PATH, "cfg")):
    logger.info("Loading model from {}".format(NEURALCOREF_MODEL_PATH))
    local_model = cached_path(NEURALCOREF_MODEL_PATH)
else:
    if not os.path.exists(NEURALCOREF_MODEL_PATH):
        os.makedirs(NEURALCOREF_MODEL_PATH)
    logger.info("Getting model from {} or cache".format(NEURALCOREF_MODEL_URL))
    downloaded_model = cached_path(NEURALCOREF_MODEL_URL)

    logger.info("extracting archive file {} to dir {}".format(downloaded_model, NEURALCOREF_MODEL_PATH))
    with tarfile.open(downloaded_model, 'r:gz') as archive:
        archive.extractall(NEURALCOREF_CACHE)

def add_to_pipe(nlp, **kwargs):
    coref = NeuralCoref(nlp.vocab, **kwargs)
    nlp.add_pipe(coref, name='neuralcoref')
    return nlp



"""
Utilities for working with the local dataset cache.
This file is adapted from the AllenNLP library at https://github.com/allenai/allennlp
Copyright by the AllenNLP authors.
"""
from __future__ import (absolute_import, division, print_function, unicode_literals)

import json
import logging
import os
import shutil
import tempfile
from functools import wraps
from hashlib import sha256
import sys
from io import open

import boto3
import requests
from botocore.exceptions import ClientError
from tqdm import tqdm

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

NEURALCOREF_CACHE = os.getenv('NEURALCOREF_CACHE', os.path.join(os.path.expanduser("~"), '.neuralcoref_cache'))

NEURALCOREF_MODEL_URL = "https://s3.amazonaws.com/models.huggingface.co/neuralcoref/neuralcoref.tar.gz"
NEURALCOREF_MODEL_PATH = os.path.join(str(NEURALCOREF_CACHE), "neuralcoref")

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def url_to_filename(url, etag=None):
    """
    Convert `url` into a hashed filename in a repeatable way.
    If `etag` is specified, append its hash to the url's, delimited
    by a period.
    """
    url_bytes = url.encode('utf-8')
    url_hash = sha256(url_bytes)
    filename = url_hash.hexdigest()

    if etag:
        etag_bytes = etag.encode('utf-8')
        etag_hash = sha256(etag_bytes)
        filename += '.' + etag_hash.hexdigest()

    return filename


def filename_to_url(filename, cache_dir=None):
    """
    Return the url and etag (which may be ``None``) stored for `filename`.
    Raise ``EnvironmentError`` if `filename` or its stored metadata do not exist.
    """
    if cache_dir is None:
        cache_dir = NEURALCOREF_CACHE

    cache_path = os.path.join(cache_dir, filename)
    if not os.path.exists(cache_path):
        raise EnvironmentError("file {} not found".format(cache_path))

    meta_path = cache_path + '.json'
    if not os.path.exists(meta_path):
        raise EnvironmentError("file {} not found".format(meta_path))

    with open(meta_path, encoding="utf-8") as meta_file:
        metadata = json.load(meta_file)
    url = metadata['url']
    etag = metadata['etag']

    return url, etag


def cached_path(url_or_filename, cache_dir=None):
    """
    Given something that might be a URL (or might be a local path),
    determine which. If it's a URL, download the file and cache it, and
    return the path to the cached file. If it's already a local path,
    make sure the file exists and then return the path.
    """
    if cache_dir is None:
        cache_dir = NEURALCOREF_CACHE

    parsed = urlparse(url_or_filename)

    if parsed.scheme in ('http', 'https', 's3'):
        # URL, so get it from the cache (downloading if necessary)
        return get_from_cache(url_or_filename, cache_dir)
    elif os.path.exists(url_or_filename):
        # File, and it exists.
        return url_or_filename
    elif parsed.scheme == '':
        # File, but it doesn't exist.
        raise EnvironmentError("file {} not found".format(url_or_filename))
    else:
        # Something unknown
        raise ValueError("unable to parse {} as a URL or as a local path".format(url_or_filename))


def split_s3_path(url):
    """Split a full s3 path into the bucket name and path."""
    parsed = urlparse(url)
    if not parsed.netloc or not parsed.path:
        raise ValueError("bad s3 path {}".format(url))
    bucket_name = parsed.netloc
    s3_path = parsed.path
    # Remove '/' at beginning of path.
    if s3_path.startswith("/"):
        s3_path = s3_path[1:]
    return bucket_name, s3_path


def s3_request(func):
    """
    Wrapper function for s3 requests in order to create more helpful error
    messages.
    """

    @wraps(func)
    def wrapper(url, *args, **kwargs):
        try:
            return func(url, *args, **kwargs)
        except ClientError as exc:
            if int(exc.response["Error"]["Code"]) == 404:
                raise EnvironmentError("file {} not found".format(url))
            else:
                raise

    return wrapper


@s3_request
def s3_etag(url):
    """Check ETag on S3 object."""
    s3_resource = boto3.resource("s3")
    bucket_name, s3_path = split_s3_path(url)
    s3_object = s3_resource.Object(bucket_name, s3_path)
    return s3_object.e_tag


@s3_request
def s3_get(url, temp_file):
    """Pull a file directly from S3."""
    s3_resource = boto3.resource("s3")
    bucket_name, s3_path = split_s3_path(url)
    s3_resource.Bucket(bucket_name).download_fileobj(s3_path, temp_file)


def http_get(url, temp_file):
    req = requests.get(url, stream=True)
    content_length = req.headers.get('Content-Length')
    total = int(content_length) if content_length is not None else None
    progress = tqdm(unit="B", total=total)
    for chunk in req.iter_content(chunk_size=1024):
        if chunk: # filter out keep-alive new chunks
            progress.update(len(chunk))
            temp_file.write(chunk)
    progress.close()


def get_from_cache(url, cache_dir=None):
    """
    Given a URL, look for the corresponding dataset in the local cache.
    If it's not there, download it. Then return the path to the cached file.
    """
    if cache_dir is None:
        cache_dir = NEURALCOREF_CACHE

    if not os.path.exists(cache_dir):
        os.makedirs(cache_dir)

    # Get eTag to add to filename, if it exists.
    if url.startswith("s3://"):
        etag = s3_etag(url)
    else:
        response = requests.head(url, allow_redirects=True)
        if response.status_code != 200:
            raise IOError("HEAD request failed for url {} with status code {}"
                          .format(url, response.status_code))
        etag = response.headers.get("ETag")

    filename = url_to_filename(url, etag)

    # get cache path to put the file
    cache_path = os.path.join(cache_dir, filename)

    if not os.path.exists(cache_path):
        # Download to temporary file, then copy to cache dir once finished.
        # Otherwise you get corrupt cache entries if the download gets interrupted.
        with tempfile.NamedTemporaryFile() as temp_file:
            logger.info("%s not found in cache, downloading to %s", url, temp_file.name)

            # GET file object
            if url.startswith("s3://"):
                s3_get(url, temp_file)
            else:
                http_get(url, temp_file)

            # we are copying the file before closing it, so flush to avoid truncation
            temp_file.flush()
            # shutil.copyfileobj() starts at the current position, so go to the start
            temp_file.seek(0)

            logger.info("copying %s to cache at %s", temp_file.name, cache_path)
            with open(cache_path, 'wb') as cache_file:
                shutil.copyfileobj(temp_file, cache_file)

            logger.info("creating metadata file for %s", cache_path)
            meta = {'url': url, 'etag': etag}
            meta_path = cache_path + '.json'
            with open(meta_path, 'w', encoding="utf-8") as meta_file:
                json.dump(meta, meta_file)

            logger.info("removing temp file %s", temp_file.name)

    return cache_path


def read_set_from_file(filename):
    '''
    Extract a de-duped collection (set) of text from a file.
    Expected file format is one item per line.
    '''
    collection = set()
    with open(filename, 'r', encoding='utf-8') as file_:
        for line in file_:
            collection.add(line.rstrip())
    return collection


def get_file_extension(path, dot=True, lower=True):
    ext = os.path.splitext(path)[1]
    ext = ext if dot else ext[1:]
    return ext.lower() if lower else ext
